"""
    *****************************************
    The Query Abstraction Layer(QAL) package.
    *****************************************

    QAL is designed to be a general library of data, data mining and database management functions.
    
    :copyright: Copyright 2010-2014 by Nicklas Boerjesson
    :license: BSD, see LICENSE for details.

"""
# Version and release information used by Sphinx for documentation and setuptools for package generation. 
__version__ = '0.4'
__release__ = '0.4.0'
__copyright__ = '2010-2014, Nicklas Boerjesson'


"""
    ************************************
    qal.dal - Database Abstraction Layer
    ************************************
    
    The goal with the DAL package is to hide the connection-related differences between the most common RDBMS database backends.
    Supported platforms are (qal.dal.types) : MySQL, PostgreSQL, Oracle, DB2 and SQL server
    
    :copyright: Copyright 2010-2014 by Nicklas Boerjesson
    :license: BSD, see LICENSE for details.
"""


"""
Created on Jan 12, 2014

@author: Nicklas Boerjesson
"""
from qal.dataset.flatfile import FlatfileDataset
from qal.dataset.xpath import XpathDataset
from qal.dataset.rdbms import RDBMSDataset
from qal.dataset.spreadsheet import SpreadsheetDataset


def dataset_from_resource(_resource):
    """Create a qal dataset from a resource"""
    try:
        if _resource.type.upper() == "FLATFILE":
            _ds =  FlatfileDataset(_resource = _resource)
        elif _resource.type.upper() == "XPATH":
            _ds =  XpathDataset(_resource = _resource)
        elif _resource.type.upper() == "RDBMS":
            _ds =  RDBMSDataset(_resource = _resource)
        elif _resource.type.upper() == "SPREADSHEET":
            _ds =  SpreadsheetDataset(_resource = _resource)
        else:
            raise Exception("qal.dataset.factory.dataset_from_resource: Unsupported source resource type: " + str(_resource.type.upper()))
    except Exception as e:
        raise Exception("qal.dataset.factory.dataset_from_resource: Failed loading resource for " + _resource.type.upper() + ".\n" + \
                        "Resource: " + str(_resource.caption)+ "(" + str(_resource.uuid) + ")\n"+ \
                        "Error: \n" + str(e))
        
    return _ds    

if __name__ == '__main__':
    pass


"""
    Helper library for QAL diff operations.
     
    :copyright: Copyright 2010-2014 by Nicklas Boerjesson
    :license: BSD, see LICENSE for details. 

"""

import difflib

# General diffutils

def diff_strings(_a,_b):
    result= "---------- String A-----------\n"
    result+= _a + "\n"
    result+= "---------- String B-----------\n"
    result+= _b + "\n"
 
    result+= "---------- Diff between A and B-----------\n" + "\n"
    for line in difflib.context_diff(_a,_b):
        result+= line 
        
    return result

def diff_files(_file_a, _file_b):
    _f_a = open(_file_a, "r")
    _f_b = open(_file_b, "r")
    _a = _f_a.read()
    _b = _f_b.read()
    _f_a.close()
    _f_b.close()
    
    return diff_strings(_a = _a, _b = _b)



def cmp_key_columns(_left, _right, _key_columns):
    """This functions compares the columns in the specified key fields only and returns data usable in a </=/>-comparer"""
    for _curr_key_column in _key_columns:

        if _left[_curr_key_column] < _right[_curr_key_column]:
            return -1
        elif _left[_curr_key_column] > _right[_curr_key_column]:
            return 1

    return 0

def match_all_columns(_left, _right):
    """Match all columns in two arrays, return false if they differ anywhere.
    
    :param array _left: The left array
    :param array _right: The right array
    
    .. note::
        This function is a *little bit* tailored to QAL needs. 
        The left column *has* to be used for column iteration, since _right might have extra columns for references to underlying structures that should not be compared.
    """
    
    # TODO: Check if this can be done more pythonically
    for _curr_column in range(0, len(_left)):
        if _left[_curr_column] != _right[_curr_column]:
            return False

    return True   

def compare(_left, _right, _key_columns, _full):
    """ The compare function takes two structurally identical 2-dimensional matrices,
        _left and _right, matches them using the columns in _key_colums,
        and returns a tuple of the results.
        
        :param 2d-array _left: The left matrix
        :param 2d-array _right: The right matrix
        :param array _key_columns: An array with the key columns
        :param bool _full: If the _full parameter is True, also the values in the rows are compared, and the third result, _different is populated with a list of rows where the values differ.
        
        
        :return array _missing_left: rows present in _right, but not in _left
        :return array _missing_right: rows present in _left, but not in _right
        :return array _difference: rows that by keys are present in _left and _right, but differ with in rows
        :return 2d-array _right_s: The _right matrix, but sorted by keys, often useful when one wants to continue massaging the data.
        
        .. note::
            The results are not in original order, but sorted by their keys.
    """    
    _missing_left = []
    _missing_right = []
    _difference = []

    # Order _left and _right using key columns
    try:
        if len(_key_columns) == 1:
            _left_s = sorted(_left, key=lambda d: ( d[_key_columns[0]]))
            _right_s = sorted(_right, key=lambda d: ( d[_key_columns[0]]))
        elif len(_key_columns) == 2:
            _left_s = sorted(_left, key=lambda d: ( d[_key_columns[0]], d[_key_columns[1]]))
            _right_s = sorted(_right, key=lambda d: ( d[_key_columns[0]], d[_key_columns[1]]))
        elif len(_key_columns) == 3:
            _left_s = sorted(_left, key=lambda d: ( d[_key_columns[0]], d[_key_columns[1]], d[_key_columns[2]]))
            _right_s = sorted(_right, key=lambda d: ( d[_key_columns[0]], d[_key_columns[1]], d[_key_columns[2]]))
        elif len(_key_columns) == 0:
            raise Exception("Error in compare, at least one key column is required.")
        else:
            raise Exception("Err..sorry, only 3 key columns are supported currently, too tired to make it dynamic. :-)")
    except TypeError as e:
        if str(e).find("TypeError: unorderable types"):
            raise Exception("There seem to be data of different types in the same column.\n"
                            "Perhaps data need to be cast to some common data type, for example string. \n"
                            "Error:" + str(e))
        else:
            raise Exception(str(e))


    # From top, loop data sets, compare all rows
    _left_idx = _right_idx = 0
    _left_len = len(_left_s)
    _right_len = len(_right_s)
    while _left_idx < _left_len and _right_idx < _right_len:
        #print("_left_idx :" + str(_left_idx) + " value: "+str(_left_s[_left_idx][_key_columns[0]]) + " | _right_idx : "  + str(_right_idx)+ " value: "+str(_right_s[_right_idx][_key_columns[0]]))
        _cmp_res = cmp_key_columns(_left_s[_left_idx], _right_s[_right_idx], _key_columns)
        #print("_cmp_res :" + str(_cmp_res))
        if _cmp_res < 0:
            #print("_missing_right.append " + str(_left_s[_left_idx]))
            _missing_right.append([_left_idx,_right_idx,_left_s[_left_idx]])
            _left_idx+= 1
        elif _cmp_res > 0:
            #print("_missing_left.append " + str(_right_s[_right_idx]))
            _missing_left.append([_left_idx,_right_idx, _right_s[_right_idx]])            
            _right_idx+= 1
        else:
            # Keys are the same and _full is set, check all data 

            if _full == True and match_all_columns(_left_s[_left_idx], _right_s[_right_idx]) != True:
                # Differing columns found, add _row to _difference
                _difference.append([_left_idx, _right_idx, _left_s[_left_idx], _right_s[_right_idx]])
            _left_idx+= 1
            _right_idx+= 1
            
    # Add remainders to missing
    if _left_idx < _left_len:   
        for _curr_item in _left_s[_left_idx: _left_len]:
            # print("_missing_right.append (post) " + str([_left_idx, len(_missing_right) + 1, _curr_item]))
            _missing_right.append([_left_idx, _left_idx, _curr_item])

    if _right_idx < _right_len:
        for _curr_item in _right_s[_right_idx: _right_len]:
            # print("_missing_left.append (post)" + str([len(_missing_left) + 1, _right_idx,_curr_item]))
            _missing_left.append([_right_idx, _right_idx,_curr_item])

    return _missing_left, _missing_right, _difference, _right_s


def diff_to_text(_missing_left, _missing_right, _different):
    """Creates a textual representation of the differences
    
        .. note::
            Not implemented.
    
    """
    _diff_text = ""
        
    return _diff_text





"""
    Functions for handling lists.
    
    :copyright: Copyright 2010-2014 by Nicklas Boerjesson
    :license: BSD, see LICENSE for details.
"""

def CI_index(_List, _value):
    """Case-insensitively finds an item in a list"""
    if _value != None:
        for index in range(len(_List)):
            if _List[index].lower() == _value.lower():
                return index
    
    return -1;

def unenumerate(value, _Type):
    """Returns the value of a specific type""" 
    return value[_Type]   

def find_next_match(_list, _start_idx, _match):
    """Finds next _match from _start_idx"""
    for _curr_idx in range(_start_idx, len(_list)):
        if _list[_curr_idx] == _match:
            return _curr_idx
    return -1

def find_previous_match(_list, _start_idx, _match):
    """Finds previous _match from _start_idx"""
    for _curr_idx in range(_start_idx, 0, -1):
        if _list[_curr_idx] == _match:
            return _curr_idx
    return -1

def pretty_list(_array):
    """Returns proper array data representation syntax, but with each row on a new text row, to make it more readable and usable."""
    
    def _handle_types(_data):
        if isinstance(_data, list):
            return str(_data)
        else:
            if isinstance(_data, str):
                return "['" + _data + "']"    
            else:
                return "[" + str(_data) + "]"       
    
    _result = ""
    if _array is None:
        return False
    elif len(_array)==1:
        return "[\n" + _handle_types(_array[0]) + "\n]"
    elif len(_array) > 0:

        for _row_idx in range(len(_array) - 1) :
            _result+= _handle_types(_array[_row_idx]) + ",\n"

        _result+= _handle_types(_array[_row_idx + 1]) + "\n"
        return "[\n" + _result + "]"
    else:
        return "[]"
    
    
        



"""
    Access functionality for resources.
    
    :copyright: Copyright 2010-2014 by Nicklas Boerjesson
    :license: BSD, see LICENSE for details. 
"""
from urllib.parse import unquote

from qal.common.xml_utils import XML_Translation

from lxml import etree
import os


def add_xml_subitem(_parent, _nodename, _nodetext):
    _curr_item = etree.SubElement(_parent, _nodename)
    _curr_item.text = _nodetext
    return _curr_item


def resource_types():
    """Returns a list of the supported resource types"""
    return ["CUSTOM", "FLATFILE", "MATRIX", "XPATH", "RDBMS"]

class Resource(object):
    """The resource class represents a QAL resource. 
    Could be any entity like a database server, flat file or a web page.
    Resources have a globally unique resource uuid"""
    
    uuid = None
    """The unique identifier of the resources. Used to access the resource, wherever its data is stored."""
    type = None
    """The type of the resources. Possible current values: "CUSTOM", "FLATFILE", "MATRIX", "XPATH", "RDBMS"."""
    caption = None
    """The "friendly" name of the resource. For example "Database BDD04"."""
    base_path = None
    """If this resource was read from an XML, base path is the directory where the XML was located.
    Makes it possible to deduce the absolute path from relative paths in the XML files, making them slightly more portable."""
    data = None
    """A dict of all the custom data that belongs to the resource. Access this way: _filename = data['file_name'].'"""
    def __init__(self):
        self.uuid = None
        self.type = None
        self.caption = None
        self.base_path = None
        self.data = {}


        
    def as_xml_node(self):
        """This function encode an XML structure into resource objects. Uses lxml as opposed to parse_xml()."""
        _resource = etree.Element("resource")
        _resource.set("caption", self.caption)
        _resource.set("type", self.type)
        _resource.set("uuid", self.uuid)
        # Loop data. Sorted to be predictable enough for testing purposes
        for _curr_data_key, _curr_data_value in sorted(self.data.items()):
            if _curr_data_value is not None:
                if isinstance(_curr_data_value, list):
                    _item_parent = add_xml_subitem(_resource, _curr_data_key, "")
                    for _item in _curr_data_value:
                        add_xml_subitem(_item_parent, "item", str(_item))
                else:
                    add_xml_subitem(_resource, _curr_data_key, str(_curr_data_value))

        
        return _resource
    

class Resources(XML_Translation):
    '''
    The resource class provides access to resources available through either XML-definitions or callback functions.
    '''
    

    local_resources = None
    """Local list of resources"""
    external_resources_callback = None
    """Callback method for external lookup"""

    base_path = None
    """If resource was loaded from a file, its path is stores here, useful when translating relative paths."""


    def __init__(self, _resources_node=None, _resources_xml=None, _external_resources_callback=None, _base_path = None):
        '''
        The argument _resources_node is an XML node from which local resources are parsed. 
        The argument _external_resources_callback is a user provided callback function 
        that has the same arguments as the get_resource-function.  
        '''

        self.local_resources = dict()

        if _base_path != None:
            self.base_path = os.path.dirname(_base_path)
        else:
            self.base_path = None

        if _resources_node != None or _resources_xml != None:
            self.parse_xml(_resources_node, _resources_xml)




    def get_resource(self, _uuid):
        """Returns the resource with the corresponding uuid"""
        
        _resource = None
        
        """Check local list"""
        if self.local_resources and _uuid in self.local_resources:
            _resource = self.local_resources[_uuid]
        
        """Lookup externally"""
        if (_resource == None) and (self.external_resources_callback):
            _resource = self.external_resources_callback(_uuid)
        
        if _resource == None:
            raise Exception("get_resource: Resource not found - uuid: " + _uuid)
        else:
            return _resource
     
    
    
    def parse_xml(self, _resources_node=None, _resources_xml=None):
        """Parses an XML structure into resource objects. Uses lxml."""


        if _resources_node == None:
            _parser = etree.XMLParser(remove_blank_text=True)
            _root_node = etree.fromstring(_resources_xml, _parser) 
            if _root_node.tag != "resources":
                _resources_node = _root_node.find("resources")
            else:
                _resources_node = _root_node            

        if _resources_node.base:
            self.base_path = os.path.dirname(unquote(_resources_node.base.replace('file:///', '')))
        
        self.local_resources = dict()
        
        for _curr_resource_node in _resources_node.findall("resource"):
            
            if _curr_resource_node.get("uuid") is not None:
                self._debug_print("parse_xml: Create new resource object")
                _new_resource = Resource()
                _new_resource.uuid = _curr_resource_node.get("uuid")
                _new_resource.type = _curr_resource_node.get("type")
                _new_resource.caption = _curr_resource_node.get("caption")
                _new_resource.base_path = self.base_path
                
                for _curr_resource_data in _curr_resource_node.findall("*"):
                    _new_data = []
                    if len(_curr_resource_data.findall("item")) > 0:
                        for _curr_item in _curr_resource_data.findall("item"):
                            _new_data.append(_curr_item.text)
                        _new_resource.data[str(_curr_resource_data.tag).lower()] = _new_data
                        self._debug_print("parse_xml: Add datas "+ str(_curr_resource_data.tag).lower() + " " +  str(_new_resource.data[str(_curr_resource_data.tag).lower()]) , 1)
                    else:
                        _new_resource.data[str(_curr_resource_data.tag).lower()] = _curr_resource_data.text
                        self._debug_print("parse_xml: Add data "+ str(_curr_resource_data.tag).lower() + " " + str(_new_resource.data[str(_curr_resource_data.tag).lower()]) , 1)
                        
            
                self.local_resources[_new_resource.uuid] = _new_resource
                self._debug_print("parse_xml: Append resource: "+_new_resource.caption + " uuid: " + _new_resource.uuid + " type: " + _new_resource.type  , 4)

    def as_xml_node(self):
        """This function encode resources structure into an XML structure."""
        _xml_node = etree.Element("resources")
        # Loop resources. Sorted to be predictable enough for testing purposes
        for _curr_resource_key, _curr_resource_value in sorted(self.local_resources.items()):
            _xml_node.append(_curr_resource_value.as_xml_node())
            
        return _xml_node


'''
Created on Sep 19, 2010

@author: Nicklas Boerjesson
@note: This module accesses the Parameter_* and Verb_* classes 
from the SQL.py and generates meta data by analyzing its objects.
'''

# Import the entire SQL namespace to make it searchable and suppress warnings.
from qal.sql.sql import * #@UnusedWildImport #IGNORE:W0401
from qal.sql.base import *
from qal.sql.remotable import *
from qal.dataset.flatfile import FlatfileDataset #@UnusedWildImport #IGNORE:W0401
from qal.dataset.xpath import XpathDataset #@UnusedWildImport #IGNORE:W0401
from qal.dataset.matrix import MatrixDataset #@UnusedWildImport #IGNORE:W0401

def list_parameter_classes():
    """List all parameter classes"""
    result = list()
    for k in globals().items():
        if (k[0][0:9]).lower() == 'parameter':
            result.append(k[0]) 
    return result    
def list_verb_classes():
    """List all verb classes"""
    result = list()
    for k in globals().items():
        if (k[0][0:4]).lower() == 'verb':
            result.append(k[0]) 
    return result    

def list_class_properties(_classname):
    """List properties for the class specified in _classname"""
    _classname = _classname.rpartition(".")[2]
    result = list()
    if (_classname in globals()):
        for k in globals()[_classname].__dict__.items():
            if not hasattr(k[1], '__call__') and k[0][0:1] != '_': 
                result.append(k[0])
        _mro_length = len(globals()[_classname]().__class__.__mro__)        
        if _mro_length > 2:
            for _mro_idx in range(1, _mro_length - 1):
                result.extend(list_class_properties(globals()[_classname]().__class__.__mro__[_mro_idx].__name__))
    
    return result

def find_class(_name, _raiserror = True):
    """Using name, find a class reference""" 
    try:
    
        Obj = None
        ObjName = None
        for k in globals().items():
            if k[0].lower() == _name.lower():
                Obj = k[1]()
                ObjName = k[0]
                break
        # Python types
        if _name == "list":
            return [], "list"
    except Exception as e:
        raise Exception("sql_meta - find_class: Error looking for " + str(_name) + " : " + str(e))
    
    if (Obj == None and _raiserror):
        raise Exception("sql_meta - find_class: Cannot find matching class - " + _name)
    
    return Obj, ObjName
    
                



'''
Created on Jan 8, 2012

@author: Nicklas Boerjesson
'''

from qal.dal.types import DB_DB2,DB_ORACLE, DB_POSTGRESQL
from qal.sql.utils import db_specific_object_reference
from datetime import date, datetime
from qal.dataset.custom import CustomDataset

class MatrixDataset(CustomDataset):
 
    """The matrix dataset holds a two-dimensional array of data"""
    
    dal = None
    
    def __init__(self):
        '''
        Constructor
        '''
        super(MatrixDataset, self ).__init__()
        
        
        
    def load(self):
        """Load data. Not implemented as it is not needed in the matrix descendant"""
        pass
    
    def save (self):
        """Save data. Not implemented as it is not needed in the matrix descendant"""
        pass
    
    def as_sql(self, _db_type):
        """Generate SQL"""
        # TODO: Review this code, as it is almost copied from the custom class. 
        #TODO: Add check here and in custom so that it is impossible to not fill all field_names when used.
        _result = []
        _add_field_names = (len(self.data_table) > 0) and (len(self.field_names) == len(self.data_table[0]))
        for _row in self.data_table:
            _curr_row = []
            for _col_idx in range(len(_row)):
                _str_col = ''
                _col = _row[_col_idx]
                if (_col == None):
                    _str_col = 'NULL'                    
                elif (isinstance(_col, bool)):
                    if (_col == True):
                        if (_db_type == DB_POSTGRESQL):
                            _str_col = 'TRUE'
                        else:
                            _str_col = '\'1\''
                    elif (_col == False):
                        if (_db_type == DB_POSTGRESQL):
                            _str_col = 'FALSE'
                        else:
                            _str_col = '\'0\''
                elif ((isinstance(_col, int)) or isinstance(_col, float)):
                    _str_col = str(_col)
                elif ((isinstance(_col, date)) or isinstance(_col, date)):
                    _str_col = "'" + _col.isoformat(' ') + "'"
                else:
                    _str_col = "'" + str(_col) + "'"
            
                if (_add_field_names):                    
                    _curr_row.append(_str_col + ' AS ' + db_specific_object_reference(self.field_names[_col_idx], _db_type))
                else:
                    _curr_row.append(_str_col)
                    
            _add_field_names = False
            
            if (_db_type == DB_DB2):
                _result.append("SELECT " + ",".join(_curr_row) +' FROM sysibm.sysdummy1')
            elif (_db_type == DB_ORACLE):
                _result.append("SELECT " + ",".join(_curr_row) +' FROM DUAL')
            else:
                _result.append("SELECT " + ",".join(_curr_row))
            
        return str("\nUNION\n".join(_result))

        


"""
Created on Sep 12, 2013

@author: Nicklas Boerjesson
"""


""" 
This module contains all the base classes for the SQL-related code.
The reason is simply to make things more manageable and split things up a bit.
"""

from qal.sql.types import DEFAULT_ROWSEP




class ParameterBase(object):
    """This class is a base class for all parameter classes."""
    _row_separator = DEFAULT_ROWSEP
    """The default row separator for SQL generation."""
    _parent = None
    """The parent of the object"""
    _base_path = None
    """The base path of the XML file from which it was loaded, used for relative paths."""

    def __init__(self, _row_separator = None):
        super(ParameterBase, self ).__init__()
        if _row_separator != None: 
            self._row_separator = _row_separator
            
    def _generate_sql(self, _db_type): 
        """Generate SQL for specified database engine"""
        raise Exception(self.__class__.__name__ + "._generate_sql() is not implemented")
    
    def as_sql(self, _db_type):
        
        if hasattr(self,'resource_uuid') and self.resource_uuid and self._check_need_prepare():
            return self._bring_into_context(_db_type)
        else:
            return self._generate_sql(_db_type)
        
        
        
class SqlList(list):
    """This is the base class for lists of class instances."""    

    def __init__(self, _itemclasses = None):
        super(SqlList, self ).__init__()
        if _itemclasses != None:
            self._itemclasses = _itemclasses

    def as_sql(self, _db_type): 
        """Generate SQL for specified database engine"""
        result = '' 
        for _item in self:
#            TODO: Check if the below list functionality is ever needed/wanted.             
#            result+= DEFAULT_ROWSEP + '--****************' + str(self.itemclasses) + ' number ' + index +'****************************' + DEFAULT_ROWSEP 
            if hasattr(_item, 'as_sql'):
                result+= _item.as_sql(_db_type)
            else:
                result+= _item

        return result  
    
    
   
class ParameterExpressionItem(ParameterBase):
    """The superclass of all classes that can be considered part of an expression"""
    operator = 'C'  
    
    def __init__(self,_operator = None):
        super(ParameterExpressionItem, self ).__init__()
        if _operator != None:
            self.operator = _operator
        else:
            self.operator = 'C' 
    




"""
    ***********************************************************
    qal.sql.sql holds the class structure representation of SQL
    ***********************************************************
    .. note:: 
        * The mySQL DDL implementation defaults to using innoDB, since stuff like foreign keys and other very important security features are lacking from myISAM.
        * All ParameterBase descendants property names are named in a specific way, so that one from that name can discern what types are allowed. For example: sources means that it is a list of ParameterSource.
        * Parameter_* means that it is some form of input, Verb_* means that this statement can be executed stand alone.
        * No parameters can be required in either Verb_* or Parameter_* classes. If they are, the classes cannot be inspected by qal.sql.meta.list_class_properties.


    :warning: Changes and new classes must satisfy both the import/export of data structures and schema generation.
    :copyright: Copyright 2010-2014 by Nicklas Boerjesson
    :license: BSD, see LICENSE for details.
"""

from qal.sql.types import DEFAULT_ROWSEP, expression_item_types, tabular_expression_item_types
from qal.dal.types import DB_POSTGRESQL, DB_MYSQL, DB_ORACLE, DB_DB2  # , DB_SQLSERVER

from qal.sql.base import ParameterBase, SqlList, ParameterExpressionItem
from qal.sql.remotable import ParameterRemotable

from qal.sql.utils import add_operator, parenthesise, oracle_add_escape, add_comma, make_operator, \
    check_for_param_content, \
    none_as_sql, error_on_blank, comma_separate, make_function, db_specific_operator, db_specific_object_reference, \
    citate, check_not_null, curr_user, db_specific_datatype, curr_datetime, add_comma_rs, oracle_create_auto_increment, \
    handle_temp_table_ref, datatype_to_parameter
from qal.sql.types import condition_part
from qal.dataset.flatfile import FlatfileDataset
from qal.dataset.xpath import XpathDataset
from qal.dataset.matrix import MatrixDataset
from qal.dal.types import DB_SQLSERVER
from qal.common.strings import make_path_absolute


class ParameterExpression(ParameterExpressionItem):
    """Holds an expression"""

    expressionitems = None

    def __init__(self, _expressionitems=None, _operator=None):
        super(ParameterExpression, self).__init__(_operator)

        if _expressionitems != None:
            self.expressionitems = _expressionitems
        else:
            self.expressionitems = SqlList()

    def _generate_sql(self, _db_type):
        """Generate SQL for specified database engine"""
        result = ''
        # Loop expressions
        for index, item in enumerate(self.expressionitems):
            # Some want || instead of + for concatenation
            if (_db_type == DB_POSTGRESQL or _db_type == DB_DB2):
                result += add_operator(index, make_operator(item.operator, True)) + item.as_sql(_db_type)
            else:
                result += add_operator(index, make_operator(item.operator, False)) + item.as_sql(_db_type)
        if len(self.expressionitems) > 1:
            return parenthesise(result)
        else:
            return result


class ParameterString(ParameterExpressionItem):
    """Holds a string parameter/value"""

    string_value = ''
    escape_character = ''

    def __init__(self, _string_value='', _operator=None, _escape_character=None):
        super(ParameterString, self).__init__(_operator)
        self.string_value = _string_value
        if (_escape_character != None):
            self.escape_character = _escape_character
        else:
            self.escape_character = ''


    def _generate_sql(self, _db_type):
        """Generate SQL for specified database engine"""
        if _db_type != DB_ORACLE:
            return "'" + self.string_value + "'"
        else:
            return oracle_add_escape("'" + self.string_value + "'", self.escape_character)


class ParameterNumeric(ParameterExpressionItem):
    """Holds a numeric parameter/value"""
    numeric_value = ''

    def __init__(self, _numeric_value='', _operator=None):
        super(ParameterNumeric, self).__init__(_operator)
        self.numeric_value = _numeric_value

    def _generate_sql(self, _db_type):
        """Generate SQL for specified database engine"""

        # Check if it is a valid float
        try:
            float(self.numeric_value)
        except ValueError:
            if (str(self.numeric_value).lower() != 'null') and (check_for_param_content(self.numeric_value) == False):
                raise Exception("A numeric parameter must be numeric. Value:" + self.numeric_value)
        return str(self.numeric_value)
        # TODO: Check if any flavors allow other decimal signs.


class ParameterParameter(ParameterExpressionItem):
    """Holds a parameter to be used in prepared statements."""

    datatype = ''

    def __init__(self, _datatype='', _operator=None):
        super(ParameterParameter, self).__init__(_operator)
        self.datatype = _datatype


    def _generate_sql(self, _db_type):
        """Generate SQL for specified database engine"""

        # This is temporary, awaiting proper PyMySQL support for %d.
        return datatype_to_parameter(_db_type, self.datatype)


class ParameterIn(ParameterExpressionItem):
    """Hold an SQL IN-statement"""
    in_values = None

    def __init__(self, _in_values='', _operator=None):
        super(ParameterIn, self).__init__(_operator)
        self.in_values = _in_values

    def _generate_sql(self, _db_type):
        """Generate SQL for specified database engine"""

        # Handle strings separately. Somewhat non-pythonic, but the alternative would be far more complicated.
        if (type(self.in_values) is ParameterString):
            return "IN " + parenthesise(self.in_values.string_value)
        else:
            return "IN " + parenthesise(self.in_values.as_sql(_db_type))


class ParameterDataset(ParameterExpressionItem, ParameterRemotable):
    """Holds a dataset from an external, non-SQL source"""
    data_source = None


    def __init__(self, _data_source=None):


        if (_data_source != None):
            self.data_source = _data_source
        else:
            self.data_source = None

        self.resource_uuid = None

    def dataset_to_sql(self):
        """Generate SQL for specified database engine """

        # TODO: Is this used?
        if (self._dataset.loaded):

            _SQL = ''

            print("in dataset_to_sql" + str(self._dataset))

            return _SQL
        else:
            raise Exception('Dataset_To_SQL: Dataset not loaded.');

    def _generate_sql(self, _db_type):
        if (self.data_source):

            """Handle relative filename"""
            if hasattr(self.data_source, 'filename'):
                self.data_source.filename = self.data_source.filename
                self.data_source._base_path = self._base_path

            self.data_source.load()
            return "(" + self.data_source.as_sql(_db_type) + ")"
        else:
            raise Exception('ParameterDataset.as_sql : data_source not set.');


class ParameterIdentifier(ParameterExpressionItem):
    """Holds an identifier(column-, table or other reference)"""
    identifier = ''
    prefix = ''

    def __init__(self, _identifier='', _operator=None, _prefix=None):
        super(ParameterIdentifier, self).__init__(_operator)
        self.identifier = _identifier
        if _prefix != None:
            self.prefix = _prefix
        else:
            self.prefix = ''


    def _generate_sql(self, _db_type):
        """Generate SQL for specified database engine"""
        # Add prefix if set
        if self.prefix != '':
            _tmp_prefix = self.prefix + '.'
        else:
            _tmp_prefix = ''
        # Add quote if needed
        if _db_type in (DB_POSTGRESQL, DB_DB2, DB_ORACLE):
            return _tmp_prefix + '"' + str(handle_temp_table_ref(self.identifier, _db_type)) + '"'
        else:
            return _tmp_prefix + str(handle_temp_table_ref(self.identifier, _db_type))


class ParameterCast(ParameterExpressionItem):
    """A Cast() converts an expression to a specified datatype.
    Properties:
    * expression is a list of expression items(sql_types.expression_item_types()).
    * datatype is a string containing the datatype(as defined in sql_types.data_types())

    Example: http://192.168.0.210/mediawiki/index.php/Unified_BPM_DatabaseAbstractionLayer#ParameterCast.28ParameterExpressionItem.29"""

    expression = ''
    datatype = ''

    def make_cast(self, _value):
        """Generate the CAST syntax"""
        return str('CAST' + parenthesise(_value + ' AS ' + self.datatype))

    def __init__(self, _expression=None, _datatype=None, _operator=None):
        super(ParameterCast, self).__init__(_operator)
        if _expression != None:
            self.expression = _expression
        else:
            self.expression = SqlList(expression_item_types())

        self.datatype = _datatype

    def _generate_sql(self, _DB_Type):
        """Generate SQL for specified database engine"""
        return self.make_cast(self.expression.as_sql(_DB_Type))


class ParameterFunction(ParameterExpressionItem):
    """Holds an SQL function call"""
    parameters = None
    name = ''

    def __init__(self, _parameters=None, _name='', _operator=None):
        super(ParameterFunction, self).__init__(_operator)

        if _parameters != None:
            self.parameters = _parameters
        else:
            self.parameters = SqlList(expression_item_types())

        self.name = _name


    def _generate_sql(self, _db_type):
        """Generate SQL for specified database engine"""
        result = ''
        for index, item in enumerate(self.parameters):
            result += add_comma(index, item.as_sql(_db_type))
        return make_function(self.name, result)


class ParameterWhen(ParameterBase):
    """Holds a WHEN statement"""
    conditions = None
    result = None

    def __init__(self, _conditions=None, _result=None):
        super(ParameterWhen, self).__init__()
        self.conditions = _conditions
        self.result = _result

    def _generate_sql(self, _db_type):
        """Generate SQL for specified database engine"""
        return 'WHEN ' + self.conditions.as_sql(_db_type) + ' THEN ' + self.result.as_sql(_db_type)


class ParameterCase(ParameterExpressionItem):
    """Holds a CASE statement (see ParameterWhen)"""
    when_statements = None
    else_statement = None

    def __init__(self, _when_statements=None, _else_statement=None, _operator=None):
        super(ParameterCase, self).__init__(_operator)

        if _when_statements != None:
            self.when_statements = _when_statements
        else:
            self.when_statements = SqlList(expression_item_types())

        self.else_statement = _else_statement

    def _generate_sql(self, _DB_Type):
        """Generate SQL for specified database engine"""
        result = 'CASE'
        for item in self.when_statements:
            result += ' ' + item.as_sql(_DB_Type)
        if self.else_statement != None:
            result += ' else_statement ' + self.else_statement.as_sql(_DB_Type)
        result += ' END'
        return result

        return result


class ParameterSet(ParameterBase):
    """This class holds a set.
    In SQL, that means more than one tabular datasets is combined using a set operator like UNION."""
    # expression could be any table-valued expression
    subsets = None
    set_operator = None

    def __init__(self, _subsets=None, _set_operator=None):
        super(ParameterBase, self).__init__()

        # One can either have tabular source data, or an SQL expression 

        if _subsets != None:
            self.subsets = _subsets
        else:
            self.subsets = SqlList(tabular_expression_item_types())
        if _set_operator != None:
            self.set_operator = _set_operator
        else:
            self.set_operator = None

    def _generate_sql(self, _db_type):
        """Generate SQL for specified database engine"""
        _sqls = []
        # Loop all expressions into a list. 
        [_sqls.append(none_as_sql(x, _db_type, '')) for x in self.subsets]
        # Separate them with operators.
        return ('\n' + self.set_operator + '\n').join(_sqls)


class ParameterSource(ParameterBase, ParameterRemotable):
    """This class hold a source of data that can be used with a FROM or JOIN-statement."""
    # expression could be any table-valued expression
    expression = None

    conditions = None
    alias = ''
    join_type = None

    def __init__(self, _expression=None, _conditions=None, _alias='', _join_type=None):
        super(ParameterSource, self).__init__()

        # One can either have tabular source data i union:ed sets, a sub select, an SQL identifier or a noSQL data set. 

        if _expression != None:
            self.expression = _expression
        else:
            self.expression = SqlList(expression_item_types())

        if _conditions != None:
            self.conditions = _conditions
        else:
            self.conditions = ParameterConditions()

        self.alias = _alias

        self.join_type = _join_type
        self.resource_uuid = None

    def _generate_sql(self, _db_type):
        """Generate SQL for specified database engine"""
        _sql = none_as_sql(self.expression, _db_type, '')

        return _sql


class ParameterOrderByItem(ParameterExpression):
    """This class holds an order by-statement"""
    # Sort direction, ASC or DESC
    direction = None

    def __init__(self, _expressionitems=None, _direction=None):
        super(ParameterOrderByItem, self).__init__(_expressionitems)
        self.direction = _direction

    def _generate_sql(self, _db_type):
        """Generate SQL for specified database engine"""
        return super(ParameterOrderByItem, self)._generate_sql(_db_type) + " " + self.direction


class VerbSelect(ParameterExpressionItem, ParameterRemotable):
    """This class holds a SELECT statement. """
    fields = None
    sources = None
    order_by = None
    top_limit = None

    # Internal variables
    _post_verb = ''
    _post_sql = ''

    # TODO: Implement GROUP BY and HAVING-support.
    # GROUP_BY = ''
    #HAVING = ''

    def __init__(self, _fields=None, _sources=None, _operator=None, _order_by=None):
        super(VerbSelect, self).__init__(_operator)

        if _fields != None:
            self.fields = _fields
        else:
            self.fields = SqlList("ParameterIdentifier")

        if _sources != None:
            self.sources = _sources
        else:
            self.sources = SqlList("ParameterSource")

        if _order_by != None:
            self.order_by = _order_by
        else:
            self.order_by = SqlList("ParameterExpressionItem")

        self.top_limit = None
        self.resource_uuid = None

    def add_limit(self, _db_type):
        """Generate SQL for specified database engine for limits on number of rows (TOP/LIMIT/FETCH FIRST)"""
        if self.top_limit != None and int(self.top_limit) > 0:
            if (_db_type in [DB_MYSQL, DB_POSTGRESQL]):
                self._post_sql = 'LIMIT ' + str(int(self.top_limit))
            elif (_db_type in [DB_DB2]):
                self._post_sql = 'FETCH FIRST ' + str(int(self.top_limit)) + ' ROWS ONLY '
                # Oracles' solution is not handled here, but in the as_sql-WHERE handling.
            elif (_db_type != DB_ORACLE):
                self._post_verb = 'TOP ' + str(int(self.top_limit)) + ' '


    def _generate_sql(self, _db_type):
        """Generate SQL for specified database engine"""

        if len(self.sources) > 0:
            self.add_limit(_db_type)

        # Add select and its fields
        result = 'SELECT ' + self._post_verb
        if (self.fields != None) and (len(self.fields) > 0):
            for index, item in enumerate(self.fields):
                result += add_comma(index, item.as_sql(_db_type))
        else:
            result += '*'


        # Add FROM
        if len(self.sources) > 0:
            result += ' FROM '
            result += self.sources[0].as_sql(_db_type)
            if self.sources[0].alias != None and self.sources[0].alias != '':
                if (_db_type != DB_ORACLE):
                    result += ' AS ' + self.sources[0].alias
                else:
                    result += ' ' + self.sources[0].alias

        # DB2 needs a dummy FROM, even if there is no source table. 
        elif (_db_type == DB_DB2):
            result += ' FROM sysibm.sysdummy1 '
        # Oracle, too, needs a dummy FROM, even if there is no source table. 
        elif (_db_type == DB_ORACLE):
            result += ' FROM dual '

        # Add joins
        if len(self.sources) > 1:
            for index, item in enumerate(self.sources):

                if index > 0:
                    """As as_sql is not called, as_sql has to be called explicitly."""
                    item.as_sql(_db_type)
                    if item.join_type:
                        result += ' ' + item.join_type
                    result += ' JOIN ' + none_as_sql(item.expression, _db_type,
                                                     _error='VerbSelect: Joins must contain a statement or a reference to a table.')
                    result += ' AS ' + error_on_blank(item.alias, 'VerbSelect: Joins must have aliases.')
                    if item.join_type != "CROSS":
                        result += ' ON ' + none_as_sql(item.conditions, _db_type,
                                                       _error='VerbSelect: Joins must have conditions.')

        # Add WHERE 
        if len(self.sources) > 0:
            _num_conds = len(self.sources[0].conditions)
            if _num_conds > 0:
                result += ' WHERE ' + self.sources[0].conditions.as_sql(_db_type)

            # Add Oracles TOP/LIMIT solution
            # TODO: Check out: Would this ever be applicable without sources?
            if (_db_type == DB_ORACLE):
                _num_conds = len(self.sources[0].conditions)
                if _num_conds > 0:
                    result += ' AND '
                else:
                    result += ' WHERE '
                #This is a *very* ugly solution, but the only way I could find out.
                if (self.top_limit > 0):
                    result += '(ROWNUM < ' + str(int(self.top_limit) + 1) + ')'

        if len(self.order_by) > 0:
            orderresult = ''
            for currItem in self.order_by:
                if orderresult == '':
                    orderresult = ' ORDER BY '
                else:
                    orderresult += ', '

                orderresult += currItem.as_sql(_db_type)
            result += orderresult

        if self._post_sql != '':
            result += DEFAULT_ROWSEP + self._post_sql

        self._post_sql = ''
        self._post_verb = ''

        return result

    def append_field_identifier(self, _identifier):
        """Helper to append field identifier classes to field list only using names"""
        _ident = ParameterIdentifier(_identifier)

        self.fields.append(_ident)


class ParameterCondition(ParameterBase):
    """This class holds a condition, that is a comparison (IF A=B)"""
    left = None
    right = None
    operator = ''
    and_or = ''

    def __init__(self, _left=None, _right=None, _operator='', _and_or=''):
        super(ParameterCondition, self).__init__()
        if _left != None:
            self.left = _left
        else:
            self.left = SqlList(condition_part())
        if _right != None:
            self.right = _right
        else:
            self.right = SqlList(condition_part())

        self.operator = _operator
        self.and_or = _and_or

    def as_sql(self, _db_type, _index=0):
        """Generate SQL for specified database engine (index if for handling when it is the first in a list of conditions)"""
        """Note that the ParameterBase.as_sql is not called, this is because 1) Conditions are in context 2) Needs _index argument"""
        # TODO: Handle ILIKE for PostgreSQL.
        if (_index != 0):
            error_on_blank(self.and_or,
                           'ParameterCondition: Conditions must be separated by a logical operator. "and_or" is not set.')
            _result = ' ' + self.and_or + ' '

        else:
            _result = ''

        _result += parenthesise(
            self.left.as_sql(_db_type) + ' ' + db_specific_operator(self.operator, _db_type) + ' ' + self.right.as_sql(
                _db_type))
        return _result


class ParameterAssignment(ParameterBase):
    """This class holds a assignment, that is assign the result of the expression to the right to the identifier on the left."""
    left = None
    right = None


    def __init__(self, _left=None, _right=None):
        super(ParameterAssignment, self).__init__()
        if _left != None:
            self.left = _left
        else:
            self.left = None
        if _right != None:
            self.right = _right
        else:
            self.right = None

    def as_sql(self, _db_type):
        """Generate SQL for specified database engine"""
        # TODO: Handle ILIKE for PostgreSQL.

        _result = self.left.as_sql(_db_type) + ' = ' + self.right.as_sql(_db_type)
        return _result


class ParameterConditions(SqlList):
    """This class holds a list of condition or list of conditions ((A=B AND C=D) OR (E=F))"""

    def __init__(self):
        super(ParameterConditions, self).__init__(["ParameterCondition", "ParameterConditions"])


    def get_first_and_or(self):
        """Return the first and/or to know if it should add its own
        (If it is the first condition there is no point in adding its or, but rather the parent conditions' operator."""
        if (len(self)) > 0:
            _first_item = self[0]
            if (hasattr(_first_item, "get_first_and_or")):
                return _first_item.get_first_and_or()
            else:
                return ' ' + _first_item.and_or + ' '
        else:
            raise "ParameterConditions: Invalid structure - Cannot get and_or operator from empty list of conditions."


    def as_sql(self, _db_type, _parent_index=0):
        """Generate SQL for specified database engine"""
        _result = ''

        for _index, _item in enumerate(self):
            _result += _item.as_sql(_db_type, _index)

        _result = parenthesise(_result)

        if (_parent_index != 0):
            _result = self.get_first_and_or() + _result

        return _result


class ParameterField(ParameterBase):
    """Holds a field definition (SELECT _FIELD1 AS _FIELD) """
    expression = None
    alias = ''

    def __init__(self, _expression=None, _alias=''):
        super(ParameterField, self).__init__()
        if _expression == None:
            self.expression = SqlList(expression_item_types())
        else:
            self.expression = _expression
        self.alias = _alias

    def _generate_sql(self, _DB_Type):
        """Generate SQL for specified database engine"""
        if self.alias != '':
            return self.expression.as_sql(_DB_Type) + ' AS ' + self.alias
        else:
            return self.expression.as_sql(_DB_Type)


class ParameterConstraint(ParameterBase):
    """Hold a key constraint declaration"""
    name = ''
    constraint_type = None
    references = None  # First key is considered local in a FK.
    checkconditions = None

    def __init__(self, _name='', _constraint_type=None, _references=None, _checkconditions=None):
        super(ParameterConstraint, self).__init__()
        self.name = _name
        self.constraint_type = _constraint_type
        if _references != None:
            self.references = _references
        else:
            self.references = SqlList()

        if _checkconditions != None:
            self.checkconditions = _checkconditions
        else:
            self.checkconditions = SqlList(["ParameterCondition", "ParameterConditions"])


    def _generate_sql(self, _db_type):
        """Generate SQL for specified database engine"""
        result = 'CONSTRAINT ' + db_specific_object_reference(self.name, _db_type) + ' ' + self.constraint_type
        if (self.constraint_type == "CHECK"):
            result += ' ' + self.checkconditions.as_sql(_db_type)
        if (self.constraint_type == "FOREIGN KEY"):
            result += ' ' + parenthesise(self.references[0].as_sql(_db_type)) + ' REFERENCES ' + citate(
                self.references[1].identifier, _db_type) + parenthesise(self.references[2].as_sql(_db_type))
        if (self.constraint_type == "PRIMARY KEY"):
            result += ' ' + parenthesise(comma_separate(self.references, _db_type))
        if (self.constraint_type == "UNIQUE"):
            result += ' ' + parenthesise(comma_separate(self.references, _db_type))
        if (self.constraint_type == "DEFAULT"):
            result += ' ' + parenthesise(self.references[0].as_sql(_db_type))
        return result


class VerbCreateIndex(ParameterBase):
    """Holds a statement for creating database indices"""
    name = ''
    index_type = None
    tablename = ''
    columnnames = None


    def __init__(self, _name='', _index_type=None, _tablename='', _columnnames=None):
        super(VerbCreateIndex, self).__init__()
        self.name = _name
        self.index_type = _index_type
        self.tablename = _tablename
        if _columnnames != None:
            self.columnnames = _columnnames
        else:
            self.columnnames = SqlList('string')

    def _generate_sql(self, _db_type):
        """Generate SQL for specified database engine"""
        check_not_null("VerbCreateIndex",
                       [[self.name, "name"], [self.index_type, "index_type"], [self.tablename, "tablename"]])
        # Handle DB2s strange deviation #1
        if (_db_type == DB_DB2 and (self.index_type == "CLUSTERED" or self.index_type == "NONCLUSTERED")):
            result = 'CREATE INDEX ' + db_specific_object_reference(self.name, _db_type) + chr(13)
        else:
            result = 'CREATE ' + self.index_type + ' INDEX ' + db_specific_object_reference(self.name, _db_type) + chr(
                13)

        result += 'ON ' + self.tablename + '('
        for index, item in enumerate(self.columnnames):
            result += add_comma(index, db_specific_object_reference(item, _db_type))
        result += ')'
        # Handle DB2s strange deviation #2
        if (_db_type == DB_DB2 and self.index_type == "CLUSTERED"):
            result += chr(13) + 'CLUSTER'
        return result


class ParameterColumndefinition(ParameterBase):
    """Holds a physical table column definition (not to confused with ParameterField which is a reference to one) """
    name = ''
    datatype = ''
    notnull = None
    default = ''

    def __init__(self, _name='', _datatype='', _notnull=None, _default=''):
        super(ParameterColumndefinition, self).__init__()
        self.name = _name
        self.datatype = _datatype
        self.notnull = _notnull
        self.default = _default


    def _generate_sql(self, _db_type, _mysql_pk=False):
        """Generate SQL for specified database engine"""
        result = db_specific_object_reference(self.name, _db_type) + ' ' + db_specific_datatype(self.datatype, _db_type)
        if (_db_type == DB_MYSQL and _mysql_pk == True and self.datatype.lower() == 'serial'):
            result += ' PRIMARY KEY'

        if (self.notnull == True and _db_type != DB_ORACLE ):
            result += ' NOT NULL'

        if (self.default != ''):

            tmp_default = self.default
            tmp_default = tmp_default.replace('::curruser::', curr_user(_db_type))
            tmp_default = tmp_default.replace('::currdatetime::', curr_datetime(_db_type))
            if _db_type not in (DB_MYSQL, DB_DB2):
                result += ' DEFAULT ' + parenthesise(tmp_default)
            else:
                result += ' DEFAULT ' + tmp_default

        if (self.notnull == True and _db_type == DB_ORACLE ):
            result += ' NOT NULL'

        if (self.notnull == False):
            result += ' NULL'

        return result


class ParameterDDL(ParameterBase):
    """Parent class for SQL DDL(Data Definition Language) statement classes"""
    _post_sql = ''

    def __init__(self, _operator=None):
        super(ParameterDDL, self).__init__()
        self._post_sql = ''


class VerbCreateTable(ParameterDDL):
    """Holds a CREATE TABLE statement"""
    name = ''
    columns = None
    constraints = None
    _post_statements = None

    def __init__(self, _name=None, _columns=None, _constraints=None):
        super(VerbCreateTable, self).__init__()
        if _name != None:
            self.name = _name
        else:
            self.name = ''

        if _columns != None:
            self.columns = _columns
        else:
            self.columns = SqlList("ParameterColumndefinition")

        if _constraints != None:
            self.constraints = _constraints
        else:
            self.constraints = SqlList(["ParameterConstraint", "ParameterConstraints"])

        self._post_statements = list()


    def get_post_statements(self):
        """Return post statements. That is statements run after the main SQL"""
        return self._post_statements

    def make_columns(self, _db_type):
        """Generate DDL for all columns"""
        result = ''
        for index, item in enumerate(self.columns):
            if (item.datatype.lower() == 'serial'):
                if (_db_type == DB_ORACLE):
                    self._post_statements.append(oracle_create_auto_increment(self, item))
                if (_db_type == DB_MYSQL and len(
                        self.constraints) == 0):  # TODO: This might have to really look for primary key constraints to be safe.
                    result += add_comma_rs(index, item._generate_sql(_db_type, True), self._row_separator)
                else:
                    result += add_comma_rs(index, item.as_sql(_db_type), self._row_separator)
            else:
                result += add_comma_rs(index, item.as_sql(_db_type), self._row_separator)
        return result

    def make_constraints(self, _db_type):
        """Generate SQL for all constraints"""
        result = ''
        for index, item in enumerate(self.constraints):
            result += add_comma_rs(index, item.as_sql(_db_type), self._row_separator)
        return result


    def _generate_sql(self, _db_type):
        """Generate SQL for specified database engine"""
        self._post_statements = []
        if self.name[0] == "#":
            if _db_type in [DB_POSTGRESQL, DB_MYSQL]:
                _result = "CREATE TEMPORARY TABLE "
            elif _db_type == [DB_DB2]:
                _result = "DECLARE GLOBAL TEMPORARY TABLE "
            elif _db_type == [DB_ORACLE]:
                _result = "CREATE GLOBAL TEMPORARY TABLE "
            else:
                _result = "CREATE TABLE "

            _result += citate(self.name[1:len(self.name)], _db_type) + " (" + self._row_separator
        else:
            _result = "CREATE TABLE "
            _result += citate(self.name, _db_type) + " (" + self._row_separator

        _result += self.make_columns(_db_type)
        if (len(self.constraints) > 0):
            _result += ',' + self._row_separator + self.make_constraints(_db_type)
        _result += self._row_separator + ')'

        if _db_type == DB_MYSQL:
            _result += ' ENGINE=InnoDB'

        return _result


class VerbInsert(ParameterBase):
    """This class holds an INSERT statement"""

    destination_identifier = None
    column_identifiers = None
    data = None

    def __init__(self, _destination_identifier=None, _column_identifiers=None, _select=None):
        super(VerbInsert, self).__init__()
        if _destination_identifier != None:
            self.destination_identifier = _destination_identifier
        else:
            self.destination_identifier = None

        if _column_identifiers != None:
            self.column_identifiers = _column_identifiers
        else:
            self.column_identifiers = SqlList("ParameterIdentifier")

        if _select != None:
            self.data = _select
        else:
            self.data = None

    def makeIdentifiers(self, _db_type):
        """Generate SQL for the identifiers"""
        result = ''
        for currIndex, currIdent in enumerate(self.column_identifiers):
            if currIndex > 0:
                result = result + ', '
            result = result + currIdent.as_sql(_db_type)
        return result

    def _generate_sql(self, _db_type):
        """Generate SQL for specified database engine"""

        if len(self.column_identifiers) > 0:
            result = 'INSERT INTO ' + self.destination_identifier.as_sql(_db_type) + ' (' + self.makeIdentifiers(
                _db_type) + ')' + DEFAULT_ROWSEP
            # Add data and remove parenthesis. Removing them here could certainly be seen as ugly but the reason 
            # is that MySQL cannot handle them. It would break everything else any other way.
            if self.data:
                tmpSQL = self.data.as_sql(_db_type)
                if isinstance(self.data, ParameterSet):
                    result = result + tmpSQL
                else:
                    result = result + tmpSQL.lstrip('(').rstrip(')')
        else:
            raise Exception('VerbInsert.as_sql: No column_identifiers specified!')

        return result


class VerbUpdate(ParameterBase):
    """This class holds an INSERT statement"""

    table_identifier = None
    assignments = None
    conditions = None

    def __init__(self, _table_identifier=None, _assignments=None, _conditions=None):
        super(VerbUpdate, self).__init__()
        if _table_identifier != None:
            self.table_identifier = _table_identifier
        else:
            self.table_identifier = None

        if _assignments != None:
            self.assignments = _assignments
        else:
            self.assignments = SqlList("ParameterAssignment")

        if _conditions != None:
            self.conditions = _conditions
        else:
            self.conditions = ParameterConditions()

    def _generate_assignments_sql(self, _db_type):
        _results = []
        for _curr_assignment in self.assignments:
            _results.append(_curr_assignment.as_sql(_db_type))

        return ", ".join(_results)

    def _generate_sql(self, _db_type):
        """Generate SQL for specified database engine"""

        if len(self.assignments) > 0:
            _result = 'UPDATE ' + self.table_identifier.as_sql(_db_type) + DEFAULT_ROWSEP
            _result += 'SET' + DEFAULT_ROWSEP + self._generate_assignments_sql(_db_type) + DEFAULT_ROWSEP
            _result += 'WHERE ' + self.conditions.as_sql(_db_type)
        else:
            raise Exception('.as_sql(_db_type) + DEFAULT_ROWSEP.as_sql: No column_identifiers specified!')

        return _result


class VerbDropTable(ParameterBase):
    """This class holds a DELETE statement"""
    name = None
    """
       The name of the table to be dropped.
    """


    def __init__(self, _name=None):
        super(VerbDropTable, self).__init__()

        if _name != None:
            self.name = _name
        else:
            self.name = None


    def _generate_sql(self, _db_type):
        """Generate SQL for specified database engine"""
        return 'DROP TABLE ' + citate(self.name, _db_type)


class VerbDelete(ParameterBase):
    """This class holds a DELETE statement"""
    sources = None
    """
       @warning: Important! The first source must have a ParameterIdentifier as expression that specifies the target table.
    """


    def __init__(self, _sources=None, _operator=None):
        super(VerbDelete, self).__init__()

        if _sources != None:
            self.sources = _sources
        else:
            self.sources = SqlList("ParameterSource")


    def _generate_sql(self, _db_type):
        """Generate SQL for specified database engine"""

        result = ''
        # Add FROM
        if len(self.sources) > 0:
            # Extract target table from expression in first source.
            if (len(self.sources[0].expression) > 0 and
                    isinstance(self.sources[0].expression[0], ParameterIdentifier)):
                if (_db_type == DB_POSTGRESQL):
                    result += 'DELETE FROM ' + self.sources[0].expression[0].as_sql(_db_type) + ' ' + self.sources[
                        0].alias + DEFAULT_ROWSEP
                else:
                    result += 'DELETE ' + self.sources[0].expression[0].as_sql(_db_type) + DEFAULT_ROWSEP

            else:
                raise Exception("Error from VerbDelete.as_sql: Could not find identifier in first source expression.")

            if (_db_type != DB_POSTGRESQL):
                result += 'FROM  '
                result += self.sources[0].as_sql(_db_type)
                if self.sources[0].alias != '':
                    if (_db_type != DB_ORACLE):
                        result += ' AS ' + self.sources[0].alias
                    else:
                        result += ' ' + self.sources[0].alias

        # Add JOINS/USING
        if len(self.sources) > 1:
            for index, item in enumerate(self.sources):
                if index > 0:
                    if (_db_type != DB_POSTGRESQL):
                        result += ' JOIN ' + none_as_sql(item.expression, _db_type,
                                                         _error='VerbDelete: Joins must contain a statement or a reference to a table.')
                    else:
                        result += ' USING ' + none_as_sql(item.expression, _db_type,
                                                          _error='VerbDelete(for Postgresql): Joins must contain a statement or a reference to a table.')
                    result += ' AS ' + error_on_blank(item.alias, 'VerbDelete: Joins must have aliases.')
                    if (_db_type != DB_POSTGRESQL):
                        result += ' ON ' + none_as_sql(item.conditions, _db_type,
                                                       _error='VerbDelete: Joins must have conditions.')

        # Add WHERE 
        if len(self.sources) > 0:
            _num_conds = len(self.sources[0].conditions)
            if (_db_type == DB_POSTGRESQL) and (len(self.sources) > 1):
                _num_conds += len(self.sources[1].conditions)
            if (_num_conds > 0):
                result += ' WHERE '
                if len(self.sources[0].conditions) > 0:
                    result += self.sources[0].conditions.as_sql(_db_type)
                if (_db_type == DB_POSTGRESQL) and (len(self.sources) > 1) and (len(self.sources[1].conditions) > 0):
                    result += ' ' + self.sources[1].conditions.as_sql(_db_type)
                    if len(self.sources) > 2:
                        raise Exception(
                            'VerbDelete: To be able to generalize functionality, only one join is allowed due to Postgresql proprietary DELETE FROM .. USING syntax.')

        return result

    def append_field_identifier(self, _identifier):
        """Helper function to add ParameterIdentifier instances using just the field names"""
        _ident = ParameterIdentifier(_identifier)

        self.fields.append(_ident)


class VerbCustom(ParameterDDL):
    """This class holds custom statements (written, not generated) for all platforms.
    This is for when what is currenctly implementet do not suffice"""

    sql_mysql = ''
    sql_postgresql = ''
    sql_oracle = ''
    sql_db2 = ''
    sql_sqlserver = ''

    def _generate_sql(self, _db_type):
        """Return the specified SQLs for each database engine"""
        return [self.sql_mysql, self.sql_postgresql, self.sql_oracle, self.sql_db2, self.sql_sqlserver][_db_type]

    def __init__(self):
        super(VerbCustom, self).__init__()
        self.sql_mysql = ''
        self.sql_postgresql = ''
        self.sql_oracle = ''
        self.sql_db2 = ''
        self.sql_sqlserver = ''
    
        


'''
Created on Sep 20, 2010

@author: Nicklas Boerjesson
@note: This module defines the basic types used in SQL. There are also some helper functions.
'''


"""Constants"""

DEFAULT_ROWSEP = chr(13)

from csv import list_dialects
from qal.dataset.xpath import xpath_data_formats

"""
 Global types:
 All type declarations have a enumeration array that must be kept up to date!
"""
def constraint_types():
    """Returns a list of the supported constraint types"""
    return ['NOT NULL', 'UNIQUE', 'PRIMARY KEY', 'FOREIGN KEY', 'CHECK', 'DEFAULT']
    

def index_types():
    """Returns a list of the supported index types"""
    return ['UNIQUE', 'CLUSTERED', 'NONCLUSTERED'];

def quoting_types():
    """Returns a list of the supported quoting modes"""
    return ['QUOTE_MINIMAL', 'QUOTE_ALL', 'QUOTE_NONE']
        
def data_types():
    """Returns a list of the supported data types"""
    # @note: string(3000) is added to give some leeway for DB2s default table page size of 4000. 
    # TODO: Investigate BLOB support
    return ['integer', 'string', 'string(255)', 'string(3000)', 'float', 'serial', 'timestamp', 'boolean']

def data_source_types():
    """Returns a list of the supported data source types"""
    return ['FlatfileDataset', 'XpathDataset', 'MatrixDataset']

def boolean():
    """Returns a list of the supported boolean values"""
    return ['true', 'false']

def and_or():
    """Returns a list of the supported logical operators"""
    return ['AND', 'OR']

def set_operator():
    """Returns a list of the supported set operators"""
    return ['UNION', 'INTERSECT', 'DIFFERENCE']

def join_types():
    """Returns a list of the supported join types"""
    return ['INNER','LEFT OUTER', 'RIGHT OUTER', 'FULL OUTER', 'CROSS']

def expression_item_types():
    """Returns a list of the supported expression types""" 
    return ['VerbSelect','ParameterExpression',
                'ParameterString','ParameterNumeric',
                'ParameterIdentifier','ParameterCast',
                'ParameterFunction', 'ParameterIn', 'ParameterDataset', 'ParameterCase', 'ParameterSet']
    
def tabular_expression_item_types(): 
    """Returns a list of the supported tabular expression types""" 
    return ['VerbSelect','ParameterDataset', 'ParameterSet']

def in_types():
    """Returns a list of what is supported in a IN-statement""" 
    return ['VerbSelect', 'ParameterString']

        
    
def condition_part(): 
    """Returns a list of the supported condition parts""" 
    return ['ParameterConditions','ParameterCondition', 'ParameterExpression'] + expression_item_types()


def sql_property_to_type(_property_name):
    """Translates a property name to a type, like decimal or string.
    Property names in the SQL.py class structure are chosen to not collide."""
    
    _property_name = _property_name.lower()
    
    # Basic types
    
    if _property_name in ['name','default','tablename','alias','default','operator',\
                        'identifier','escape_character','string_value','operator','sql_mysql',\
                        'sql_postgresql','sql_oracle','sql_db2','sql_sqlserver','row_separator',\
                        'prefix', 'direction', 'operator','table', 'parameters', 'delimiter',\
                        'filename', 'target_table', 'resource_uuid', 'temporary_table_name',\
                        'temporary_table_name_prefix', 'rows_xpath', "quotechar", "skipinitialspace"]:
        return ['string']
    
    elif _property_name == 'numeric_value':
        return ['decimal']

    elif _property_name in ['notnull', 'has_header']:
        return ['boolean']    
    
    elif _property_name in ['top_limit']:
        return ['integer']    
    
    elif _property_name == 'quoting':
        return ['quoting', quoting_types()]

    # Simple types
    
    elif _property_name == 'datatype':
        return ['datatypes', data_types()]
    
    elif _property_name == 'set_operator':
        return ['set_operator', set_operator()]
    
    elif _property_name == 'join_type':
        return ['join_type', join_types()]

    elif _property_name == 'data':
        return ['tabular_expression_item', tabular_expression_item_types()]
    
    elif _property_name == 'subsets':
        return ['Array_tabular_expression_item', tabular_expression_item_types()]    

    elif _property_name == 'and_or':
        return ['and_or', and_or()]
    
    elif _property_name == 'constraint_type':
        return ['constraint_types', constraint_types()]
    
    elif _property_name == 'index_type':
        return ['index_types', index_types()]
    
    elif _property_name == 'xpath_data_format':
        return ['xpath_data_format',xpath_data_formats()]
    
    elif _property_name in ['expression','parameters','result','else_statement','orderby','expressionitems']:
        return ['Array_expression_item', expression_item_types()]
    
    elif _property_name in ['in_values']:
        return ['in_types', in_types()]   
    
    elif _property_name in ['data_source']:
        return ['data_source_types', data_source_types()]   

    elif _property_name in ['left','right']:
        return ['condition_part', condition_part()]
    
    # Complex types
    elif _property_name in ['checkconditions','conditions']:
        return ['Array_ParameterCondition']

    elif _property_name in ['columnnames']:
        return ['Array_string']
        
    elif _property_name == 'columns':
        return ['Array_ParameterColumndefinition']
    
    elif _property_name in ['destination_identifier', 'table_identifier']:
        return ['ParameterIdentifier']

    elif _property_name in ['column_identifiers','references']:
        return ['Array_ParameterIdentifier']
        
    elif _property_name == 'constraints':
        return ['Array_ParameterConstraint']
    
     
    elif _property_name == 'fields':
        return ['Array_ParameterField']

    elif _property_name == 'select':
        return ['VerbSelect']
     
    elif _property_name == 'sources':
        return ['Array_ParameterSource']
    
    elif _property_name == 'csv_dialect':
        return ['file_types', list_dialects()] 

    
    elif _property_name == 'when_statements':
        return ['Array_ParameterWhen']
    elif _property_name == 'order_by':
        return ['Array_ParameterOrderByItem']
    elif _property_name == 'assignments':
        return ['Array_ParameterAssignment']
    else:
        raise Exception("sql_property_to_type: Unrecognized property:" + _property_name)
    
def verbs(): 
    """Returns a list of the supported verb types""" 
    return ['VerbCreateTable','VerbCreateIndex', 'VerbSelect','VerbCustom', 'VerbInsert']
   



"""
    Helper library for XML operations and their debugging. 
    
    .. note::
        Will probably be deprecated or at least rewritten when QAL switches completely to lxml.

    :copyright: Copyright 2010-2014 by Nicklas Boerjesson
    :license: BSD, see LICENSE for details. 
"""



from urllib.request import quote, unquote
from xml.dom.minidom import Text
from qal.sql.utils import check_for_param_content
from xml.dom.minidom import Document, parseString

def xml_set_cdata(_node, _value, _lowercase=False):
    """Helper to set character data in an XML tree"""
    if _value != None and _value != "":
        sec = Text()
        if _value is str:
            _value = quote(_value)
        if _lowercase: # Force lowercase.
            sec.data =  _value.lower() 
        else:
            sec.data = _value 
        
        _node.appendChild(sec)

def xml_get_text(_node):
    """Helper function to get character data from an XML tree"""
    rc = list()
    for node in _node.childNodes:
        if node.nodeType == node.TEXT_NODE:
            rc.append(node.data)
    return unquote(''.join(rc))

def xml_get_boolean(_node):
    """Helper function to get a boolean value from XSD:booleans defaults"""
    value = xml_get_text(_node)
    if value.lower() == 'true':
        return True
    elif value.lower() == 'false':
        return False
    else: 
        raise Exception('xml_get_boolean: "' + value + '" is not a boolean value')

def xml_get_numeric(_node, _type = ""):
    """Helper function to read a numeric value from the XML"""
    try:
        _value = xml_get_text(_node)
        if _type.lower() == "integer":
            number = int(_value)
        else:            
            number = float(_value)
    except:
        if str(_value).lower() == 'null':
            return 'NULL' 
        if not check_for_param_content(_value):
            raise Exception("xml_get_numeric: Invalid numeric format: " + _value)
        else:
            return _value

    return number

def xml_base_type_value(_node, _typename):
    """Using the base type name, get appropriate data"""
    if _typename in ["int", "float"]:
        return xml_get_numeric(_node)
    elif _typename in ["str"]:
        return xml_get_text(_node) 
    elif _typename in ["bool"]:
        return xml_get_boolean(_node)
    else:
        raise Exception("xml_base_type_value: Invalid base type: " + _typename)  

def xml_get_allowed_value(_node, _type):
    """Check if a value is allowed in a certain XML node"""
    value = xml_get_text(_node)

    if (value in _type[1] or value == ""):
        return value
    # Check for correct string format.
    elif value[0:8].lower() == "varchar(" and  value[8:len(value) - 1].isnumeric() and value[len(value) - 1] == ")":
        return value  
    else:
        raise Exception("xml_get_allowed_value: " + str(value) + " is not a valid value in a " + _type[0] + ". Node:" + str(_node))

def xml_find_non_text_child(_node):
    """Finds the first child that is not of the Text node type."""
    nodelist = _node.childNodes
    for node in nodelist:
        if node.nodeType != node.TEXT_NODE:
            return node
    
    return None    

def find_child_node(_node, _nodename):
    """Find all child nodes of an XML tree"""
    for node in _node.childNodes :  # visit every node <bar />
        if (node.nodeName.lower() == _nodename.lower()):
            return node
    return None    


      
class XML_Translation(object):
    """
    This is a base class for translations between XML structures and object structures. 
    """
    
    """Transcoding settings"""
    prefix_xml = 'xsi'
    prefix_schema = 'xsd'
    prefix_own = None
    
    namespace = None
    schema_uri = None
    encoding = 'utf-8'
    
    #Debugging
    debuglevel = 2
    nestinglevel = 0
    
    def __init__(self):
        '''
        Constructor
        '''
        pass
    
    """ Debugging """
    
    def _print_nestinglevel(self, _value):
        """Prints the current nesting level. Not thread safe."""
        self._debug_print(_value + ' level: ' + str(self.nestinglevel),4)
        
    def _get_up(self,_value):
        """Gets up one nesting level. Not thread safe."""
        self.nestinglevel = self.nestinglevel - 1
        self._print_nestinglevel("Leaving " + _value)

    def _go_down(self,_value):
        """Gets down one nesting level. Not thread safe."""
        self.nestinglevel = self.nestinglevel + 1
        self._print_nestinglevel("Entering " + _value)
        
    def _debug_print(self, _value, _debuglevel = 3):
        """Prints a debug message if the debugging level is sufficient."""
        if self.debuglevel >= _debuglevel:
            print(_value)
            
    """Helper utilities"""
         
    def _add_own(self, _value):
        # Adds the set SQL-prefix.
        if self.prefix_own:
            return self.prefix_own + ':' + _value
        else:
            return _value
        
    def _strip_own(self, _value):
        ps_len = len(self.prefix_own + ':')
        if _value[0:ps_len] == self.prefix_own + ':':
            return _value[ps_len:len(_value)]
        else:
            return _value
        
    def get_root_node(self, _nodename, _xml = "", _node = None ):
        
        if _node == None:
            self._debug_print(self.__class__.__name__ +".get_root_node : - XML being parsed:\n" + _xml)
            try:
                _doc = parseString(_xml)
            except Exception as e:
                raise Exception(self.__class__.__name__ +".get_root_node : Exception parsing SQL:\n" + str(e) + "\n XML: \n" + _xml)
        
            """ Find root node having _nodename."""
            for _curr_node in _doc.childNodes:
                if _curr_node.nodeName == self._add_own(_nodename):
                    return _curr_node 
  
                
            if _curr_node.nodeName != self._add_own(_nodename):
                raise Exception(self.__class__.__name__ + '.get_root_node : "' + self._add_own(_nodename) + ' top node required.')
        else:
            if _node.nodeName != _nodename and _node.nodeName != self._add_own(_nodename):
                raise Exception(self.__class__.__name__ + '.get_root_node : "' + _nodename +'" top node required.')

        return _node




def xml_isnone(_node):
    if _node == None or _node.text == None:
        return None
    else:
        return _node.text


'''
Contains functionality for reading settings from ini-files

:copyright: Copyright 2010-2014 by Nicklas Boerjesson
:license: BSD, see LICENSE for details. 
'''

import configparser

class UBPMSettings(object):
    """This class is responsible for reading settings from ini-files and holding them in memory."""
    
    global Parser
    
    def reload(self,filename):
        """Reload all information"""
        self.Parser.read(filename)
        
    def get(self, _section, _option, _default = None):
        """Get a certain option"""
        if (_default != None) and (not self.Parser.has_option(_section, _option)):
            return _default
        else:
            return self.Parser.get(_section, _option)

    
    def __init__(self, filename):
        """Constructor"""
        self.Parser = configparser.ConfigParser()
        if (filename != ''):
            # TODO: Add better config support.
            self.reload(filename)
            
        


"""
    ********************************************************************
    qal.tools is a collection of tools for data management related tasks
    ********************************************************************
    These are used by the users of QAL, like the Optimal BPM data merge/migration tool.
    (See: https://sourceforge.net/projects/optimalbpm/files/Optimal%20BPM%20tools/)
    
    There are diff, merge and transform classes, but also nmap-based network 
    discovery of RDBMS hosts and a collection of functions for querying information from them.
    
    :copyright: Copyright 2010-2014 by Nicklas Boerjesson
    :license: BSD, see LICENSE for details.
"""


"""
    ************************************
    qal.dal - Database Abstraction Layer
    ************************************
    
    The goal with DAL is to hide the connection-related differences between the most common database backends.
    
    :copyright: Copyright 2010-2014 by Nicklas Boerjesson
    :license: BSD, see LICENSE for details.
"""

from qal.dal.types import DB_MYSQL, DB_POSTGRESQL, DB_ORACLE, DB_DB2, DB_SQLSERVER, string_to_db_type, db_type_to_string
from qal.dal.conversions import parse_description, python_type_to_SQL_type
from qal.tools.discover import import_error_to_help


class DatabaseAbstractionLayer(object):
    """This class abstracts the different peculiarities of the different database backends with
    regards to connection details"""
    
    # Events
    
    on_connect = None
    """Triggered on connect"""
    # Properties
    
    connected = False
    """Is true if connected"""
    
    db_connection = None
    """The database connection"""
    db_type = None
    """Database type"""
    db_server = ''
    """Server name"""
    db_databasename = ''
    """Database name"""
    db_username = ''
    """Username"""
    db_password = ''
    """Password"""
    db_instance = ''
    """Instance"""
    db_driver = None
    """Database driver"""
    db_autocommit = True
    """Autocommit. If True, the SQL is committed immidiately, if False, commit needs to be called to commit changes."""
    db_port = None
    """The TCP port of the database server"""
    
    field_names = None
    """The field names of the dataset"""
    field_types = None
    """The field types of the fields as defined in conversions.py"""
    
    # Postgres transaction object
    _pg_xact = None
    
    def read_ini_settings(self, _ini_parser):
        """
        Read setting from the settings.
        :param _ini_parser: ConfigParser object
        """
        if _ini_parser.Parser.has_option("database", "type"):
            self.db_type        = string_to_db_type(_ini_parser.Parser.get("database", "type"))
            self.db_server      = _ini_parser.Parser.get("database", "server")   
            self.db_databasename= _ini_parser.Parser.get("database", "database_name")   
            self.db_username    = _ini_parser.Parser.get("database", "username")
            self.db_password    = _ini_parser.Parser.get("database", "password")
            self.db_port        = _ini_parser.Parser.get("database", "port")
            self.db_autocommit     = _ini_parser.get("database", "autocommit", True)
            if _ini_parser.Parser.has_option("database", "instance"):
                self.db_instance    = _ini_parser.Parser.get("database", "instance")
        else:
            print("read_ini_settings: Settings not valid, not raising error.")
            
            
    def read_resource_settings(self, _resource):
        """
        Read settings from a resource object
        :param _resource: A resource object
        """
        if _resource.type.upper() != 'RDBMS':
            raise Exception("DAL.read_resource_settings error: Wrong resource type - " + _resource.type)
        self.db_type =         string_to_db_type(_resource.data.get("db_type"))
        self.db_server =       _resource.data.get("db_server")
        self.db_databasename = _resource.data.get("db_databasename")
        self.db_instance =     _resource.data.get("db_instance")
        self.db_username =     _resource.data.get("db_username")
        self.db_password =     _resource.data.get("db_password")
        self.db_port =         _resource.data.get("db_port")
        self.db_autocommit =      _resource.data.get("db_autocommit")

    def write_resource_settings(self, _resource):
        """
        Write settings to a resource object
        :param _resource: A resource object.
        """
        _resource.type = 'RDBMS'
        _resource.data.clear()
        _resource.data["db_type"] = db_type_to_string(self.db_type)
        _resource.data["db_server"] = self.db_server
        _resource.data["db_databasename"] = self.db_databasename
        _resource.data["db_instance"] = self.db_instance
        _resource.data["db_username"] = self.db_username
        _resource.data["db_password"] = self.db_password
        _resource.data["db_port"] = self.db_port
        _resource.data["db_autocommit"] = self.db_autocommit

                       
    def connect_to_db(self):
        """Connects to the database"""
        if self.db_type == DB_MYSQL:
            try:
                import pymysql
            except ImportError as _err:
                # TODO: Add python3-mysql when available or consider mysql-connector when available for python3
                raise Exception(import_error_to_help(_module="pymysql", _err_obj=_err, _pip_package="pymysql3",
                                                     _apt_package=None, _win_package=None))

            _connection = pymysql.connect (host = self.db_server,
                            db = self.db_databasename,
                            user = self.db_username,
                            passwd = self.db_password,
                            )
            

        elif self.db_type == DB_POSTGRESQL:

            try:
                import postgresql.driver as pg_driver
            except ImportError as _err:
                raise Exception(import_error_to_help(_module="postgresql", _err_obj=_err,
                                                     _pip_package="py-postgresql",
                                                     _apt_package="python3-postgresql",
                                                     _win_package=None,
                                                     _import_comment="2014-04-16: If using apt-get, " +
                                                                     "check so version is > 1.0.3-2" +
                                                     " as there is a severe bug in the 1.02 version. " +
                                                     "See https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=724597"))

            if self.db_port == None or self.db_port == "" or self.db_port == 0:
                _port = 5432
            else:
                _port = self.db_port
            _connection = pg_driver.connect(host = self.db_server,
                                                database =  self.db_databasename, 
                                                user = self.db_username, 
                                                password = self.db_password,
                                                port = _port)
                            
        elif self.db_type in [DB_SQLSERVER, DB_DB2]:
            _connection_string = None
            try:
                import pyodbc
            except ImportError as _err:
                raise Exception(import_error_to_help(_module="pyodbc", _err_obj=_err,
                                                     _pip_package="pyodbc",
                                                     _apt_package=None,
                                                     _win_package=None,
                                                     _import_comment="(For Linux) 2014-04-16: " +
                                                                     "No apt package (python3-pyodbc)"+
                                                                     " available at this time."))
            import platform

            #TODO: Investigate if there is any more adapting needed, platform.release() can also be used.

            if self.db_type == DB_SQLSERVER:
                if platform.system().lower() == 'linux':
                    _connection_string = "DRIVER={FreeTDS};SERVER=" + self.db_server + ";DATABASE=" + \
                                         self.db_databasename +";TDS_VERSION=8.0;UID=" + self.db_username + ";PWD=" + \
                                         self.db_password + ";PORT="+str(self.db_port) + ";Trusted_Connection=no;"
                elif platform.system().lower() == 'windows':
                    _connection_string = "Driver={SQL Server};Server=" + self.db_server + ";DATABASE=" + \
                                         self.db_databasename +";UID=" + self.db_username + ";PWD=" + self.db_password +\
                                         ";PORT="+str(self.db_port) + ";Trusted_Connection=no"
                else:
                    raise Exception("connect_to_db: ODBC connections on " + platform.system() + " not supported yet.")


            elif self.db_type == DB_DB2:

                if platform.system().lower() == 'linux':
                    drivername = "DB2"
                elif platform.system().lower() == 'windows':
                    drivername = "{IBM DATA SERVER DRIVER for ODBC - C:/PROGRA~1/IBM}"
                else:
                    raise Exception("connect_to_db: DB2 connections on " + platform.system() + " not supported yet.")

                # DSN-less?{IBM DB2 ODBC DRIVER} ?? http://www.webmasterworld.com/forum88/4434.htm
                _connection_string =  "Driver=" + drivername + ";Database=" + self.db_databasename +";hostname=" + \
                                      self.db_server + ";port="+str(self.db_port) + ";protocol=TCPIP; uid=" + \
                                      self.db_username + "; pwd=" + self.db_password
                #_connection_string = "DSN=" + self.db_server + ";UID=" + self.db_username + ";PWD=" + self.db_password
                _connection = pyodbc.connect(_connection_string, autocommit=self.db_autocommit)


                # DSN-less?{IBM DB2 ODBC DRIVER} ?? http://www.webmasterworld.com/forum88/4434.htm
                _connection_string =  "Driver=" + drivername + ";Database=" + self.db_databasename +";hostname=" + \
                                      self.db_server + ";port="+str(self.db_port) + ";protocol=TCPIP; uid=" + \
                                      self.db_username + "; pwd=" + self.db_password
                #_connection_string = "DSN=" + self.db_server + ";UID=" + self.db_username + ";PWD=" + self.db_password

            print("Connect to database using connection string:  " + _connection_string)
            _connection = pyodbc.connect(_connection_string, autocommit=self.db_autocommit)
        
        # cx_Oracle in python 3.X not checked yet.
        elif self.db_type == DB_ORACLE:
            try:
                import cx_Oracle
            except ImportError as _err:
                raise Exception(import_error_to_help(_module="cx_Oracle", _err_obj=_err,
                                                     _pip_package="cx_Oracle",
                                                     _apt_package=None,
                                                     _win_package="Download and install binary .msi package from " +
                                                                  "http://cx-oracle.sourceforge.net/ and install.",
                                                     _import_comment="(Linux) 2014-04-16: No python3-pyodbc available at" +
                                                                     " build time."))

            _connection_string = self.db_username + '/' +  self.db_password + '@' + self.db_server + ':' + \
                                 str(self.db_port) + '/' + self.db_instance
            print("Connect to database using connection string:  " + _connection_string)
            _connection = cx_Oracle.connect(_connection_string)
            _connection.autocommit=self.db_autocommit
                  
        else:
            raise Exception("connect_to_db: Invalid database type.")              
      
        
        self.db_connection = _connection
        
        if self.on_connect:
            self.on_connect() 
        self.connected = True
            
        return _connection
    
    
    
    def __init__(self, _settings = None, _resource = None):
        '''
        Init
          
        '''  
        if _settings != None:      
            self.settings = _settings
            self.read_ini_settings(_settings)

        if _resource != None:
            self.resource = _resource
            self.read_resource_settings(_resource)

    def execute(self, _sql):
        """Execute the SQL statement, expect no dataset"""
        
        
        if self.db_type == DB_POSTGRESQL:
            self.db_connection.execute(_sql)
        else:
            cur = self.db_connection.cursor()
            cur.execute(_sql)
            
    def _make_positioned_params(self, _input):

            _arg_idx = 1
            _output =  _input
            
            # Simplest possible scan, prepared to handle the types differently.                        
            for _curr_idx in range(0, len(_input)):
                _chunk = _input[_curr_idx:_curr_idx +2]
                
                if _chunk == "%s":
                    _output = _output.replace(_chunk,"$" + str(_arg_idx), 1)           
                    _arg_idx+=1            
                elif _chunk == "%d":                    
                    _output = _output.replace(_chunk,"$" + str(_arg_idx), 1)           
                    _arg_idx+=1            

            return _output    
                
           
    def executemany(self, _sql, _values):
        """Execute the SQL statements , expect no dataset"""
        
       
        if self.db_type == DB_POSTGRESQL:
            # Change parameter type into Postgres positional ones, like  $1, $2 and so on.
            # TODO: Correctly handle other datatypes than string.
            _sql = self._make_positioned_params(_sql)

            _prepared = self.db_connection.prepare(_sql)
            print(_sql)
            for _row in _values:
                _prepared(*_row)
        else:
            cur = self.db_connection.cursor()
            cur.executemany(_sql, _values)
    

    def query(self, _sql):
        """Execute the SQL statement, get a dataset"""
        
        # py-postgres doesn't use the DB-API, as it doesn't work well-
        if self.db_type == DB_POSTGRESQL:
            _ps = self.db_connection.prepare(_sql)
            _res = _ps()
            if _ps.column_names != None:
                self.field_names = _ps.column_names 
                self.field_types = []                
                for _curr_type in _ps.column_types:
                    self.field_types.append(python_type_to_SQL_type(_curr_type))

        else:
            cur = self.db_connection.cursor()
            cur.execute(_sql)
            
            self.field_names, self.field_types = parse_description(cur.description, self.db_type);
            
            _res = cur.fetchall()
            
        # Untuple. TODO: This might need to be optimised, perhaps by working with the same array. 
        _results = [] 
        for _row in _res:
            _results.append(list(_row))
        
        return _results
   
    
   
    def close(self):
        """Close the database connection"""
        self.db_connection.close()
        
    def start(self):
        """Start transaction"""
        if self.db_type == DB_POSTGRESQL and self._pg_xact == None:
            self._pg_xact = self.db_connection.xact()
            self._pg_xact.start()
        else:
            self.execute("START TRANSACTION")
     
            
    def commit(self):
        """Commit the transaction"""
        if self.db_type == DB_POSTGRESQL:
            if self._pg_xact != None:
                self._pg_xact.commit()
                self._pg_xact = None
        else:
            self.db_connection.commit()

    def rollback(self):
        """Rollback the transaction"""
        if self.db_type == DB_POSTGRESQL:
            self._pg_xact.rollback()
            self._pg_xact = None
        self.db_connection.rollback()
    


"""
Created on Oct 20, 2014

@author: Nicklas Boerjesson
"""
from lxml import etree
from qal.common.substitution import Substitution
from qal.common.strings import string_to_bool, empty_when_none
from qal.common.transform import make_transformation_array_from_xml_node, make_transformations_xml_node
from qal.common.xml_utils import xml_isnone

class Mapping(object):
    is_key = None
    """If true, the mapping is a key field"""
    src_reference = None
    """A reference to the source data location within its dataset. Can be a field name, XPath or similar."""
    src_datatype = None
    """The source data data type"""
    dest_reference = None
    """A reference to the destination data location within its dataset. Can be a field name, XPath or similar."""
    substitution = None
    """An instance of the substitution class. Kept for maintaining things lite incrementors and similar."""



    def __init__(self, _xml_node=None, _substitution=None):
        """
        Constructor
        """
        self.transformations = []

        if _substitution is None:
            self.substitution = Substitution()
        else:
            self.substitution = _substitution

        if _xml_node != None:
            self.load_from_xml_node(_xml_node)



    def load_from_xml_node(self, _xml_node):
        if _xml_node != None:
            self.is_key = string_to_bool(xml_isnone(_xml_node.find("is_key")))
            self.src_reference = xml_isnone(_xml_node.find("src_reference"))
            self.src_datatype = xml_isnone(_xml_node.find("src_datatype"))
            self.dest_reference = xml_isnone(_xml_node.find("dest_reference"))
            self.transformations = make_transformation_array_from_xml_node(_xml_node.find("transformations"),
                                                                           self.substitution)

    def as_xml_node(self):

        _xml_node = etree.Element("field_mapping")
        etree.SubElement(_xml_node, "is_key").text = str(self.is_key)
        etree.SubElement(_xml_node, "src_reference").text = empty_when_none(self.src_reference)
        etree.SubElement(_xml_node, "src_datatype").text = empty_when_none(self.src_datatype)
        _xml_node.append(make_transformations_xml_node(self.transformations))
        etree.SubElement(_xml_node, "dest_reference").text = empty_when_none(self.dest_reference)

        return _xml_node



"""
Created on Oct 30, 2013

@author: Nicklas Boerjesson
"""
import platform

import sys
from qal.common.strings import empty_if_none

def platform_to_int():
    """Returns an integer depending on platform."""
    if platform.system().lower() == "linux":
        return 0
    elif platform.system().lower() == "windows":
        return 1
    elif platform.system().lower() == "darwin":
        return 2
    else:
        return 3

def import_error_to_help(_module, _err_obj, _pip_package, _apt_package, _win_package, _import_comment=None):
    """Usable to create a helpful error message if a module is missing"""
    if str(_err_obj) == "No module named '" + str(_module) + "'":
        _err_msg = "The python " + get_python_versions(_style="Minor") + " module \"" + str(_module) + "\" is not installed.\n"
        _err_msg+= ["Run sudo pip3 install " + str(_pip_package) +
                     empty_if_none(" or sudo apt-get install " + str(_apt_package), _apt_package),
                     "Either run pip install " + str(_pip_package) + empty_if_none(", download and install " + str(_win_package) +
                                                                       " or install from source.",_win_package),
                     "If available on your platform, run pip" + get_python_versions(_style="Minor") + " install " + str(_pip_package) +
                     " otherwise download and install from source."][platform_to_int()]
        return _err_msg + empty_if_none("\n" + str(_import_comment), _import_comment)
    else:
        return str(_err_obj)


def discover_database_servers(_ip):
    """Uses Nmap to scan the subnet of the specified IP-address for the following RDBMS:
    SQL Server, IBM DB2, MYSQL, Oracle and Postgres."""
    return discover_services(_ip, [1433,523,1521,5432,3306], True)


def discover_services(_ip, _ports, _verbose = False):
    """Uses Nmap to scan the subnet of the specified IP-address for services on the specified ports"""
    
    try:
        import nmap                         # import nmap.py module
    except ImportError as _err:
        raise Exception(import_error_to_help(_module= "nmap", _err_obj = _err, _pip_package = "python-nmap", _apt_package = None, _win_package = None))

    try:
        _nm = nmap.PortScanner()      # instantiate nmap.PortScanner object
    except nmap.PortScannerError:
        raise Exception('Nmap not found(nmap not installed on the system?)', sys.exc_info()[0])
    except Exception as e:
        print("Unexpected error:" + str(e), sys.exc_info()[0])
        raise Exception(str(e))
    
    _ips = _ip.split(".")
    _new_ip = ".".join(_ips[0:3])
    _port_str = ",".join(str(x) for x in _ports)
    if _verbose:
        print("Scanning " +_new_ip+".2-254, ports " + _port_str)
    
    _nm.scan(hosts=_new_ip+".2-254", arguments="-p T:" + _port_str + " -T4 -sV")


    _detected_services = []
    for _host in _nm.all_hosts():
        if _nm[_host].state() == 'up':
            if _verbose:
                print('----------------------------------------------------')
                print('Host : %s (%s)' % (_host, _nm[_host].hostname()))
                print('State : %s' % _nm[_host].state())
                
            for _proto in _nm[_host].all_protocols():
                if _verbose:
                    print('----------')
                    print('Protocol : %s' % _proto)
                
                for _port in _nm[_host][_proto]:
                    if isinstance(_nm[_host][_proto][_port], dict) and _nm[_host][_proto][_port]['state'] == 'open':
                        if _verbose:
                            print(_nm[_host].hostname() + " - " +_nm[_host][_proto][_port]['product'] + _host)
                        _detected_services.append( [_host, _nm[_host].hostname(), _nm[_host][_proto][_port]['product']]) 

    return _detected_services

def get_python_versions(_style=None):
    """
    Return the current version of python, styled in different ways:
    * major : 3
    * minor : 3.4
    * release : 3.4.4
    * build : 3.4.4 build 1300

    :param _style: Style of the the result.
    """
    _major, _minor, _release, _state, _build = sys.version_info

    if _style is None or _style.lower() == "full":
        return str(_major) + "." + str(_minor) + "." + str(_release) + " " + _state + " build " + str(_build)
    elif _style.lower() == "major":
        return str(_major)
    elif _style.lower() == "minor":
        return str(_major) + "." + str(_minor)
    elif _style.lower() == "release":
        return str(_major) + "." + str(_minor) + "." + str(_release)

    else:
        raise Exception("Error in get_python_versions: Invalid _style-parameter :'"+ _style + "'")





"""
Created on Sep 26, 2013

@author: Nicklas Boerjesson
"""

from qal.sql.sql import VerbCreateTable, VerbSelect, VerbInsert, VerbDelete, VerbUpdate, VerbDropTable, SqlList, \
    ParameterColumndefinition, ParameterIdentifier, ParameterSource

from qal.sql.utils import datatype_to_parameter


def make_column_definitions(_field_names, _field_types):
    """Based on a list of field names and field types, return a list of ParameterColumndefinition objects.

    :param _field_names: A list of field names
    :param _field_types: A list of field types
    :return: A list of ParameterColumndefinition
    """
    _columns = SqlList()
    _field_counter = 0
    _curr_column = None
    for _field_counter in range(0, len(_field_names)):
        _columns.append(
            ParameterColumndefinition(_name=_field_names[_field_counter], _datatype=_field_types[_field_counter]))
    return _columns


def create_table_skeleton(_table_name, _field_names, _field_types):
    """Creates a sql for creating tables.

    :param _table_name: The table name to use
    :param _field_names: The field(column) names
    :param _field_types: The types of the fields
    :return:
    """
    _columns = make_column_definitions(_field_names, _field_types)
    _CREATE = VerbCreateTable(_name=_table_name, _columns=_columns)
    return _CREATE


def make_column_identifiers(_field_names):
    """Create columns idenfiers from a list of field names.

    :param _field_names: A list of field names
    :return: A list of column identifies (instances of ParameterIdentifier)
    """

    _columns = SqlList()
    _field_counter = 0
    _curr_column = None
    for _field_counter in range(0, len(_field_names)):
        _columns.append(ParameterIdentifier(_identifier=_field_names[_field_counter]))
    return _columns


def make_delete_skeleton(_table_name, _key_fields):
    # TODO: Implement make_delete_skeleton

    """NOT IMPLEMENTED: Make a skeleton for an SQL DELETE statement object.

    :param _table_name:
    :param _key_fields:
    :return:
    """
    raise Exception("make_delete_skeleton isn't implemented")
    _destination_identifier = ParameterIdentifier(_identifier=_table_name)

    _column_identifiers = make_column_identifiers(_key_fields)
    return VerbDelete()


def make_update_skeleton(_table_name):
    """Make a skeleton for an SQL update object

    :param _table_name: The name of the table
    :return: An instance of VerbUpdate
    """
    _table_identifier = ParameterIdentifier(_identifier=_table_name)
    return VerbUpdate(_table_identifier=_table_identifier)


def make_insert_sql_with_parameters(_table_name, _field_names, _db_type, _field_types):
    """Make a prepared statement-type INSERT INTO ...VALUES-SQL.
    See: http://en.wikipedia.org/wiki/Prepared_statement

    :param _table_name: The table to insert into
    :param _field_names: Fields(column) names
    :param _db_type: The database backend type(qal.dal.types)
    :param _field_types: A list of field types(qal.sql.types)
    :return: Return a populated instance of an insert SQL
    """
    _destination_identifier = ParameterIdentifier(_identifier=_table_name)
    _column_identifiers = make_column_identifiers(_field_names)
    _insert = VerbInsert(_destination_identifier=_destination_identifier, _column_identifiers=_column_identifiers)
    _refs = []
    # The VALUES-part of INSERT INTO looks the same on all platforms and we don't have to care about escaping.
    for _curr_idx in range(0, len(_field_names)):
        _refs.append(datatype_to_parameter(_db_type, _field_types[_curr_idx]))

    _values_sql = "VALUES (" + ", ".join(_refs) + ")"

    return _insert.as_sql(_db_type) + _values_sql




def copy_to_table(_dal, _values, _field_names, _field_types, _table_name, _create_table=None, _drop_existing=None):
    """Copy a matrix of data into a table on the resource, return the table name.

    :param _dal: An instance of DAL(qal.dal.DAL)
    :param _values: The a list(rows) of lists(values) with values to be inserted
    :param _field_names: The name of the fields(columns)
    :param _field_types: The field types(qal.sql.types)
    :param _table_name: The name of the destination tables
    :param _create_table: Create the destination table based on _field_names, _field_types
    :param _drop_existing: If a table with the same name as the destination table already exists, drop it
    :return: The name of the destination table.
    """

    if _drop_existing == True:
        try:
            _dal.execute(VerbDropTable(_table_name).as_sql(_dal.db_type))
            _dal.commit()
        except Exception as e:
            print("copy_to_table - Ignoring error when dropping the table \"" + _table_name + "\": " + str(e))

    if _create_table == True:
        # Always create temporary table even if it ends up empty.
        _create_table_sql = create_table_skeleton(_table_name, _field_names, _field_types).as_sql(_dal.db_type)
        print("Creating " + _table_name + " table..\n" + _create_table_sql)
        _dal.execute(_create_table_sql)
        _dal.commit()

    if len(_values) == 0:
        print("copy_to_table: No source data, inserting no rows.")
    else:
        _insert_sql = make_insert_sql_with_parameters(_table_name, _field_names, _dal.db_type, _field_types)

        print("Inserting " + str(len(_values)) + " rows (" + str(len(_values[0])) + " columns)")
        _dal.executemany(_insert_sql, _values)
        _dal.commit()

    return _table_name


def select_all_skeleton(_table_name, _column_names=None):
    """Returns a "SELECT * FROM _table_name"-structure,
    if _column_names are specified, SELECT only from those columns

    :param _table_name: The name of the table to select from
    :param _column_names: Optionally, the name of the columns to select from
    :return: A populated instance of VerbSelect
    """
    _expression = ParameterIdentifier(_table_name)
    if (_column_names is None) or (_column_names == []):
        _fields = None
    else:
        _fields = make_column_identifiers(_column_names)
    _source = ParameterSource(_expression, _conditions=None, _alias=None, _join_type=None)
    _select = VerbSelect(_fields=_fields, _sources=[_source], _operator="AND")

    return _select



'''
Created on Sep 14, 2012

@author: Nicklas Boerjesson
'''

import io
from qal.common.strings import make_path_absolute

from qal.dataset.custom import CustomDataset
from qal.common.listhelper import find_next_match, find_previous_match
from lxml import _elementpath
from lxml import etree
from lxml.etree import SubElement



def xpath_data_formats():
    return ["XML", "XHTML", "HTML"] # "UNTIDY_HTML"]

class XpathDataset(CustomDataset):
    """This class implements data formats that are possible to query via XPath.
    Currently XML and HTML are implemented(XHTML isn't tested). 
    Untidy HTML will be implemented using the Beautiful Soup library.
    """

    filename = None
    """The name of the file to be loaded"""

    rows_xpath = None
    """The XPath to the "row" nodes of the data set. Ex: /books/book"""

    field_xpaths  = None
    """A list of XPaths relative to the rows_xpath leading to each field. Ex: ["title", "id"]"""

    xpath_data_format = None
    """The data format. Can be either "XML", "XHTML" or "HTML"."""

    _structure_tree = None
    """A private reference to the node tree, when load is called with _add_node_ref=True, an instance
    is kept for more efficient merging with diff data"""

    _structure_row_node_parent = None
    """A private reference to the parent node of the row nodes in the structure """     
    
    _structure_key_fields = None
    """A private list of key fields."""
    
    _structure_row_node_name = None
    """
        A private variable holding the tag name of the row node.
        
        .. todo::
            Should this be able to hold an XML?
    """

    _structure_top_node = None
    """
        A private variable holding the top node of the structure.
    """

    def __init__(self, _filename = None, _rows_xpath = None,  _resource = None):

        """Constructor.
        :param str _filename: Name of the file to parser
        :param str _rows_xpath: Path to nodes holding the rows.
        :param Resource _resource : An instance of a :py:class: qal.common.resource.resource
        """
        super(XpathDataset, self ).__init__()

        self.field_xpaths = []
        self._structure_key_fields = []
        
        if _resource:
            self.read_resource_settings(_resource)
        else:
            if _filename != None: 
                self.filename = _filename
            else:
                self.filename = None                
            if _rows_xpath != None: 
                self.rows_xpath = _rows_xpath
            else:
                self.rows_xpath = None



                


    def read_resource_settings(self, _resource):
        """Reads settings from a resource"""

        self._base_path = _resource.base_path
  
        if _resource.type.upper() != 'XPATH':
            raise Exception("XpathDataset.read_resource_settings.parse_resource error: Wrong resource type: " + _resource.type)
        self.filename   = _resource.data.get("filename")
        self.rows_xpath = _resource.data.get("rows_xpath")
        self.xpath_data_format =  _resource.data.get("xpath_data_format")
        self.field_names = _resource.data.get("field_names")
        self.field_xpaths = _resource.data.get("field_xpaths")       
        self.field_types = _resource.data.get("field_types") 
        self.xpath_text_qualifier = _resource.data.get("xpath_text_qualifier") 
        self.encoding = _resource.data.get("encoding")


    def write_resource_settings(self, _resource):
        """Write settings to a resource"""

        _resource.type = 'XPATH'
        _resource.data.clear()
        _resource.data["filename"] = self.filename
        _resource.data["rows_xpath"] = self.rows_xpath
        _resource.data["xpath_data_format"] = self.xpath_data_format
        _resource.data["field_names"] = self.field_names
        _resource.data["field_xpaths"] = self.field_xpaths
        _resource.data["field_types"] = self.field_types
        _resource.data["xpath_text_qualifier"] = self.xpath_text_qualifier
        _resource.data["encoding"] = self.encoding
              
    def _file_to_tree(self, _data_format, _reference):
        """Reads a file and chooses the right parser to make it an lxml element tree"""
        print("format_to_tree : " + _data_format)
        if _data_format == 'HTML':
            from lxml import html
            return html.parse(_reference)
        if _data_format == 'XML':
            from lxml import etree
            return etree.parse(_reference)
        else:
            raise Exception("_file_to_tree: " + _data_format + " is not supported")
        
    def _structure_parse_qal_xpath(self, _qal_xpath):
        """Parse the trailing attribute identifier from the QAL xpath string. ":" isn't allowed in xpath.
        :param str _qal_xpath: QAL xpath string. Ex. "library/books/book | library/books"
        """
        _parts = _qal_xpath.split("::")
        if len(_parts) == 2:
            return _parts[0], _parts[1]
        else:
            return _qal_xpath, None
        
        
    def load(self, _add_node_ref = None):
        """Parse file, apply root XPath to iterate over and then collect field data via field_xpaths.
            
        :param bool add_node_ref: If set, add an extra column with references to the underlying XML nodes, \
        used to efficiently apply new data.
        """
        print("Loading : " + self.filename)
        self.log_load(self.filename)  
        try:
            _tree = self._file_to_tree(self.xpath_data_format, make_path_absolute(self.filename, self._base_path))
            
        except Exception as e:
            raise Exception("XpathDataset.load - error parsing " + self.xpath_data_format + " file : " + str(e))
        
        #_root_nodes = _tree.xpath("/html/body/form/table/tr[4]/td[2]/table/tr[2]/td/table[2]/tr/td/table/tr[10]/td/table/tr")
        _root_nodes = _tree.xpath(self.rows_xpath)
        if len(_root_nodes) > 0:
            self._structure_tree = _tree
            self._structure_row_node_parent = _root_nodes[0].getparent()
            self._structure_top_node = _tree.getroot()
        
        _data = []

        if self.field_xpaths is not None and len(self.field_xpaths) > 0:
            for _curr_row in _root_nodes:
                _row_data = []
                for _field_idx in range(0, len(self.field_names)):
                    _curr_path, _curr_attribute = self._structure_parse_qal_xpath(self.field_xpaths[_field_idx])
                    if _curr_path != "":
                        _found_node = _curr_row.xpath(_curr_path)
                        if len(_found_node)>0:
                            _item_data = _found_node[0]
                        else:
                            _item_data = None
                    else:
                        _item_data = _curr_row

                    if _item_data is not None:
                        if _curr_attribute:
                            if _curr_attribute[0] == "@":
                                _row_data.append(self.cast_text_to_type(_item_data.get(_curr_attribute[1:]), _field_idx))
                            elif _curr_attribute.lower() == "xml":
                                _row_data.append(self.cast_text_to_type(_item_data.tostring(), _field_idx))
                            else:
                                raise Exception("Param_XPath.load: Error, attribute references must have the ::@AttributeName- format or ::XML. The attribute was \""+_curr_attribute + "\"")
                        else:
                            _row_data.append(self.cast_text_to_type(_item_data.text, _field_idx))
                    else:
                        _row_data.append("")


                # Add reference to XML structure.
                if _add_node_ref:
                    _row_data.append(_curr_row)

                print(str(_row_data))
                _data.append(_row_data)

            self.data_table = _data

        return _data  
    
    
    def _structure_create_xpath_nodes(self, _node, _xpath):   
        """Used an xpath to create nodes that match the path and its conditions(names, attributes and so forth)"""

        print("_create_xpath_nodes: " + str(_xpath))
        _curr_node = _node
        # Break up the string in its tokens
        _tokens =  list(_elementpath.xpath_tokenizer(_xpath))
        print(str(_tokens))   
        
        # Iterate through tokens, until we have gone through the entire XPath.
        _token_idx = 0 
        
        # But first, move past any root reference and conditions. 
        # TODO: This is a bit ugly, perhaps.
        if len(_tokens) > 1 and (_tokens[0][0] == "/") and (_tokens[1][1] == _node.tag):
            print("_create_xpath_nodes: Ignoring root node path and condition.")
            _token_idx+=1
            while _token_idx < len(_tokens) and _tokens[_token_idx][0] != "/": 
                _token_idx+=1
        
     
        while _token_idx < len(_tokens):
            print("_tokens[" + str(_token_idx) + "][0]:" + str(_tokens[_token_idx][0]))
            #Is this a new level?
            
            if _tokens[_token_idx][0] == "/":
                # Then the next is the name of the node
                _token_idx+=1                
            
            if _tokens[_token_idx][0] == "":

                _next_name = _tokens[_token_idx][1]

                #Is the next token a condition?
                
                if _token_idx + 1 < len(_tokens) and _tokens[_token_idx + 1][0] == "[" :
                    #It was, move on
                    _token_idx+=1
                    
                    #Find ending of condition
                    _end_cond_idx = self.find_next_match(_tokens, _token_idx, ("]",""))

                    if _end_cond_idx == -1: 
                        raise Exception("_create_xpath_nodes: Cannot find end of condition.\nXPath = " + _xpath) 
                    #Create relative path
                    _check_path = "".join((_a[0]+_a[1] for _a in _tokens[_token_idx -1:_end_cond_idx + 1]))

                    #Then check if it exists
                    print("_check_path:" + str(_check_path))
                    _found_nodes = _curr_node.xpath(_check_path)
                    
                    # Node found, move on
                    if _found_nodes and len(_found_nodes) == 1:
                        _token_idx+=1
                        _curr_node = _found_nodes[0]
                    else:
                        # If not found create node 
                        print("add node: " + str(_next_name)) 
                        _curr_node = SubElement(_curr_node, _next_name)
                        
                        # If they can be discerned(@id=value), add attributes
                        if _tokens[_token_idx + 3][0] == "=" and _tokens[_token_idx + 1][0] == "@":
                            print("set attribute to satisfy this path: " + str(_check_path)) 
                            _curr_node.set(_tokens[_token_idx + 2][1], str(_tokens[_token_idx + 4][0]).strip("\"'"))
                        
                    _token_idx+=5
                else:
                    if isinstance(_curr_node, str):
                        raise Exception("_structure_create_xpath_nodes - Internal error: Node is a string, missing node ref, was _add_node_ref used? \nData :" + _curr_node+" | "+_next_name+" | "+_xpath+" | "+str(_tokens))
                    _found_nodes = _curr_node.xpath(_next_name)
                    # Node found, move on
                    if len(_found_nodes) == 1:
                        _curr_node = _found_nodes[0]
                    else:
                        print("Create new node: " + str(_next_name))
                        _curr_node = SubElement(_curr_node, _next_name)

            _token_idx+=1
        return _curr_node
    
    def _structure_parse_root_path(self, _root_path):
        """Extract information from the root path to get sufficient information 
        for specifying the destination and creating new nodes. \n 
        * If many XPaths in the string, it uses the first one 
        * Parses the root node name
        * Parses the row parent node path
        * Parses the row node name"""
        
        if _root_path == "":
            raise Exception("_parse_root_path: Root path cannot be empty")
        # Parse XPath tokens
        _root_xpath_tokens =  list(_elementpath.xpath_tokenizer(self.rows_xpath))
        # Use only first path for destination
        try:
            _first_splitter = _root_xpath_tokens.index(('', '|'))
            print("_parse_root_path: Multiple XPaths, using first.")
            _split_xpath = "".join((_a[0]+_a[1] for _a in _root_xpath_tokens[0:_first_splitter]))
        except ValueError:
            _split_xpath = self.rows_xpath


        # Get the root node name
        if _root_xpath_tokens[0][0] == "/":
            _root_node_name = _root_xpath_tokens[1][1]
        else:
            raise Exception("_parse_root_path: It is necessary to have an absolute (\"/node\") top node name in the XPath")
        print("_root_node_name=" + str(_root_node_name))
        # Get the row node name

        _row_node_name_idx = find_previous_match(_root_xpath_tokens, len(_root_xpath_tokens) -1, ("/", ""))
        if _row_node_name_idx == -1:
            raise Exception("_parse_root_path: Cannot find start of condition.\nXPath = " + _root_path)
        else:
            _row_node_name = _root_xpath_tokens[_row_node_name_idx + 1][1]
            
        if _row_node_name_idx == 0:
            raise Exception("_parse_root_path: The row node cannot be the root node.\nXPath = " + _root_path)
        
        print("_structure_row_node_name=" + str(_row_node_name))
        # Move on backward, what's left is the path to the parent of the row node, we need it to be able 
        # to create rows, as the path may not return a node to find a parent from
        _row_node_parent_xpath = "".join((_a[0]+_a[1] for _a in _root_xpath_tokens[0:_row_node_name_idx])) 
        
        print("_row_node_parent_xpath=" + str(_row_node_parent_xpath))
        
        return _root_node_name, _row_node_name, _row_node_parent_xpath
    
    def _structure_init(self, _dataset):
        """Initializes the XML structure that data is to be applied to."""
        print("XpathDataset._structure_init")
        super(XpathDataset, self)._structure_init(_dataset)

        # Parse important information data from XPath 
        _root_node_name, self._structure_row_node_name, _parent_xpath = self._structure_parse_root_path(self.rows_xpath)
        
        # If the structure already loaded?
        if self._structure_row_node_parent is None:
            
            # If not try to load, or create file.    
            import os
            if os.path.exists(make_path_absolute(self.filename, self._base_path)):
                
                try:
                    self.load(_add_node_ref=True)
                except Exception as e:
                    raise Exception("XpathDataset.save - error parsing " + self.xpath_data_format + " file : " + str(e))
            else:
                # Create a tree with root node based on the first  
                
                if _root_node_name != "":
                    if self.encoding:
                        _encoding = self.encoding
                    else:
                        _encoding = "UTF-8"
                              
                    _tree = etree.parse(io.StringIO("<?xml version='1.0' ?>\n<" + _root_node_name + "/>")) 
                else:
                    raise Exception("XpathDataset.save - rows_xpath("+ str(self.rows_xpath)+") must be absolute and have at least the name of the root node. Example: \"/root_node\" ")

        # If the structure there yet? It could be an XML file with only a top node. 
        if self._structure_row_node_parent is None:
            # If not existing, create a node structure up to the parent or the row nodes
            # from the information in the xpath.
            self._structure_top_node = self._structure_create_xpath_nodes(self._structure_top_node, self.rows_xpath)
            

  
       
        
    def _structure_populate_row(self, _node, _row_data):
        for _field_idx in range(len(self.field_xpaths)):
            _curr_path, _curr_attribute = self._structure_parse_qal_xpath(self.field_xpaths[_field_idx])
            _curr_node = self._structure_create_xpath_nodes(_node, _curr_path)
            _new_value = str(_row_data[_field_idx])

            print("_curr_path      :" + str(_curr_path))
            print("_curr_attribute :" + str(_curr_attribute))
            print("_new_value :" + str(_new_value))
            #TODO: Support XML?
            if _curr_attribute:
                _curr_node.set(_curr_attribute[1:], _new_value)
            else:
                _curr_node.text = _new_value

             
    def _structure_insert_row(self, _row_idx, _row_data, _commit=True, _no_logging=None):
        if _commit is True:
            """Override parent to add XML handling"""

            # Last column is a reference
            print("_row_data: " + str(_row_data))
            _last_row_idx = len(self.data_table) -1
            if _row_idx > _last_row_idx:
                _row_node = self.data_table[_last_row_idx][len(self.data_table[_last_row_idx]) -1]
            else:
                _row_node = self.data_table[_row_idx][len(self.data_table[_row_idx]) -1]

            _new_element = etree.Element(self._structure_row_node_name)

            # Insert row node
            self._structure_row_node_parent.insert(_row_idx, _new_element)

            self._structure_populate_row(_new_element, _row_data)

        # Call parent
        super(XpathDataset, self)._structure_insert_row(_row_idx,_row_data)
        
    def _structure_update_row(self, _row_idx, _row_data, _commit=True, _no_logging=None):
        if _commit is True:
            """Override parent to add XML handling"""

            # Last column is a reference
            print("_row_data: " + str(_row_data))
            _row_node = self.data_table[_row_idx][len(self.data_table[_row_idx]) -1]


            self._structure_populate_row(_row_node, _row_data)
        
        # Call parent
        super(XpathDataset, self)._structure_update_row(_row_idx,_row_data)

    def _structure_delete_row(self, _row_idx, _commit=True, _no_logging=None):
        if _commit is True:
            """Override parent to add XML handling"""
            _row_node = self.data_table[_row_idx][len(self.data_table[_row_idx]) -1]
            self._structure_row_node_parent.remove(_row_node)
        # Call parent
        super(XpathDataset, self)._structure_delete_row(_row_idx)
        #self.data_table.pop(_row_idx)

    def save(self, _save_as = None):
        """Save the document"""
        
        if not _save_as:
            _save_as = make_path_absolute(self.filename, self._base_path())
            
        self._structure_tree.write(_save_as)
            
        self.log_save(_save_as)  
        return self._log



'''
Created on Sep 30, 2013

@author: Nicklas Boerjesson
'''
from qal.dal.dal import DatabaseAbstractionLayer

class ParameterRemotable(object):
    """This class is an auxilliary class for all parameter classes that are remotable. 
    That is, they can fetch their data from, or perform their actions at, a different location than the parent class.
    If they return data, the data will be held in the temporary table, where it can be joined with or otherwise managed.
    """
    """The temporary table name is used by owners to reference the data correctly. It is prefixed by an underscore to 
    not be shown in the external structure.(it is possible that it will be removed)"""
    _temporary_table_name = None
    """The resource object, usually loaded by the XML importer"""
    _resource = None
    """The classes' DAL connection."""
    _dal = None
    """The nearest parent that is in another database context than self"""
    _out_of_context_parent = None
    """This value identifies what context this part of the query resides in"""
    resource_uuid = None
    
            
    def _bring_into_context(self, _db_type):
        """The _bring_into_context function checks whether the resource GUID is set and if resources need fetching
        If so, it fetches the data and puts it into a dataset, which is then inserted into the parents context.
        The parent is referenced to as the out-of-context parent.
        """
       
        print(self.__class__.__name__ + "._bring_into_context: Needs preparing, preparing for resource_uuid: " + str(self.resource_uuid))
        
        if not self._resource:
            raise Exception(self.__class__.__name__ + "._bring_into_context - Error: _resource not cached")
        
        # Generate random temp table name, max 1 (#) + 8 alphanumeric characters
        import random
        import string
        char_set = string.ascii_lowercase + string.digits
        _tmp_table_name = '#' +''.join(random.sample(char_set*8,8))
       
        from qal.sql.sql import ParameterIdentifier
        
        if self._resource.type.upper() == 'RDBMS':
            if not self._dal:
                """Make connection to resource defined by the resource_uuid"""
                self._dal = DatabaseAbstractionLayer(_resource = self._resource)
                self._dal.connect_to_db()
            
            _source_sql = self._generate_sql(self._dal.db_type)
            from qal.sql.sql import ParameterSource
            if  isinstance(self, ParameterSource):
                print("is ParameterSource")
                if isinstance(self.expression[0], ParameterIdentifier):
                    
                    _source_sql = 'SELECT * FROM ' + self.expression[0].as_sql(self._dal.db_type)
                    if _tmp_table_name[1] == "#":
                        self.expression[0].identifier = _tmp_table_name[1:len(_tmp_table_name)]
                    else:
                        self.expression[0].identifier = _tmp_table_name
                    

            _data = self._dal.query(_source_sql)
            _field_names = self._dal.field_names
            _field_types = self._dal.field_types
            self._dal.close()
            
        elif self._resource.type.upper() in ["FLATFILE"]:

            from qal.dataset.flatfile import FlatfileDataset
            
            self._data_source = FlatfileDataset(_resource = self._resource)
            
            _data = self._data_source.load()
            _field_names = self._data_source.field_names
            _field_types = ["string"] * len(_field_names)
            
        elif self._resource.type.upper() in ["XPATH"]:    
            from qal.dataset.xpath import XpathDataset
            
            self._data_source = XpathDataset(_resource = self._resource)
            
            _data = self._data_source.load()
            _field_names = self._data_source.field_names
            _field_types = self._data_source.field_types
                  
        else:
            raise Exception(self.__class__.__name__ + "._bring_into_context - Error: Invalid resource type : " + str(self._resource.type))    

        
        """Does the out-of-context parent have a connection? If so, that's where the data should go."""
        if self._out_of_context_parent._dal:
            _parent_dal = self._out_of_context_parent._dal
        else:
            """If it doesn't, create one to the resource, and use that."""
            _parent_dal = DatabaseAbstractionLayer(_resource=self._out_of_context_parent._resource)
            _parent_dal.connect_to_db()
            """Also set the parents _dal, since we are using temporary tables, they need to be in the same context.""" 
            self._out_of_context_parent._dal = _parent_dal
            
        """ The sql_macros library is imported locally, to not interfere with the qal.sql.* structure."""
        from qal.sql.macros import copy_to_table             
        """Copy the data into the parents' context so the parent can access it."""
        _table_name = copy_to_table(_dal = _parent_dal, _values =_data, _field_names = _field_names, _field_types = _field_types, _table_name = _tmp_table_name, _create_table = True)
        
        _ident = ParameterIdentifier(_identifier = _table_name)
        return _ident.as_sql(_parent_dal.db_type)
        

    def _check_need_prepare(self):
        """
        Checks context, whether a call to _bring_into_context is needed. It is needed if:
        1. If the nearest parent with resource has a different resource ID
        2. If no parent has a resourceID
        
        It is NOT needed if the nearest parent with resource has the same ID. 
        
        """
        _curr_parent = self._parent
        while _curr_parent:
            print(self.__class__.__name__ + "._check_need_prepare: level up" + str(_curr_parent))
            if hasattr(_curr_parent, 'resource_uuid') and _curr_parent.resource_uuid:
                if self.resource_uuid == _curr_parent.resource_uuid:
                    if _curr_parent._dal != None: 
                        self._dal = _curr_parent
                    return False
                else:
                    print(self.__class__.__name__ + "._check_need_prepare: Different resource uuid found:" + str(_curr_parent.resource_uuid) + " (own: " +str(self.resource_uuid)+ ")")
                    self._out_of_context_parent = _curr_parent
                    return True
            _curr_parent = _curr_parent._parent

        return False
                



'''
Created on Jan 8, 2012

@author: Nicklas Boerjesson
'''
from qal.dal.conversions import python_type_to_SQL_type

from qal.dal.types import DB_DB2,DB_ORACLE, DB_POSTGRESQL
from qal.sql.utils import db_specific_object_reference
from qal.sql.sql import VerbInsert, VerbDelete, VerbUpdate, ParameterConditions
from qal.sql.sql import ParameterSource, ParameterIdentifier, ParameterParameter,ParameterCondition, ParameterAssignment
from qal.dal.dal import DatabaseAbstractionLayer
from qal.sql.macros import select_all_skeleton, create_table_skeleton
from qal.sql.base import SqlList

from qal.sql.macros import make_insert_sql_with_parameters

from qal.dataset.custom import CustomDataset

class RDBMSDataset(CustomDataset):
 
    """The RDMBS Dataset holds a two-dimensional array of data, typically representing a table in a database.
    If the data set is not a table but based on a more complex query, data will not be possible to apply to it."""
    
    dal = None
    """An instance of the Database Abstraction Layer(DAL)"""
    table_name = None
    """If set, all data in table "table_name" is loaded in its entirety into the data_table."""
    query = None
    """If set, and table_name is not, then the SQL statement contained is executed. 
    It is a text string. 

    .. todo::
    Make it possible for it to be a backend agnostic qal.sql.sql.VerbSelect object parsed from an XML structure."""

    def read_resource_settings(self, _resource):
        if _resource.type.upper() != 'RDBMS':
            raise Exception("RDBMSDataset.read_resource_settings.parse_resource error: Wrong resource type: " + _resource.type)
        
        self.dal = DatabaseAbstractionLayer(_resource = _resource)
        
        self.table_name = _resource.data.get("db_table_name")
        
        self.query = _resource.data.get("db_query")


    def write_resource_settings(self, _resource):
        """TODO: The RDBMS resource type is a special case, as it is both a database connection and/or a dataset definition.
        Should this be?"""

        if self.dal is None:
            # Clear, as the DAL is not going to
            _resource.type = 'RDBMS'
            _resource.data.clear()
        else:
            # Let the DAL go first
            self.dal.write_resource_settings(_resource)

        # Add own parameters
        _resource.data["db_table_name"] = self.table_name
        _resource.data["db_query"] = self.query

    
    def __init__(self, _resource = None):
        '''
        Constructor
        '''
        super(RDBMSDataset, self ).__init__()
        
        if _resource != None:
            self.read_resource_settings(_resource)        

    def _structure_init(self, _dataset):

        """Initializes the SQL queries that data is to be applied to, starts a database transaction"""
        # If field types aren't set, that means that the table isn't created.
        # They would have been populated as part of the loading process
        if len(self.field_types) == 0:
            if len(_dataset) == 0:
                raise Exception("RDBMSDataset.structure_init error: The _dataset cannot be empty "
                                "for field types to be derived from it.")
            try:
                self.field_types = [python_type_to_SQL_type(type(_curr_value)) for _curr_value in _dataset[0]]
                _create_table_sql = create_table_skeleton(self.table_name, self.field_names, self.field_types).as_sql(self.dal.db_type)
                print("Creating " + self.table_name + " table..\n"+_create_table_sql)
                self.dal.execute(_create_table_sql)
            except Exception as e:
                raise Exception("Merge.execute: An error occurred creating destination table: " + str(e))


        self._rdbms_init_deletes()
        if len(self._structure_key_fields) <= len(self.field_names):
            self._rdbms_init_updates()


        self._rdbms_init_inserts()
        
        self.dal.start()


        super(RDBMSDataset, self)._structure_init(_dataset)

    def _extract_data_columns_from_diff_row(self, _field_indexes, _diff_row):    
        """Extracts columns specified in _field_indexes from _diff_list"""
        _result = []

        for _curr_field in _field_indexes:
            _result.append(_diff_row[_curr_field])
        return _result        
    

    def _rdbms_init_deletes(self):
        """Generates a VerbDelete instance populated with the structural indata"""
        
        _source = ParameterSource()
        _source.expression.append(ParameterIdentifier(self.table_name))
        
        # Add the WHERE statement
        for _field_idx in self._structure_key_fields:
            _new_cond = ParameterCondition(_left = ParameterIdentifier(_identifier= self.field_names[_field_idx]),
                                            _right = ParameterParameter(_datatype = self.field_types[_field_idx]),
                                             _operator = '=', _and_or = 'AND')
            _source.conditions.append(_new_cond)
        
        # Make the VerbDelete skeleton
        _delete = VerbDelete()
        _delete.sources.append(_source)
        
        # Fetch the resource
        self._structure_delete_sql = _delete.as_sql(self.dal.db_type)


    
    def _rdbms_init_inserts(self):
        """Generates a VerbInsert instance populated with the structural indata"""

        self._structure_insert_sql = make_insert_sql_with_parameters(self.table_name, self.field_names,self.dal.db_type, self.field_types)

        
    def _rdbms_init_updates(self):
        """Generates DELETE and INSERT instances populated with the structural indata """


        # Add assignments to all fields except the key fields and add conditions for all key fields.
                
        _field_names_ex_keys = []
        _field_names_ex_keys_datatypes = []
        _key_field_names = []
        _key_field_datatypes = []
        
        # Create lists of field names and types excluding and including keys
        for _curr_field_idx in range(len(self.field_names)):
            if _curr_field_idx in self._structure_key_fields:
                _key_field_names.append(self.field_names[_curr_field_idx])
                _key_field_datatypes.append(self.field_types[_curr_field_idx])
            else:
                _field_names_ex_keys.append(self.field_names[_curr_field_idx])
                _field_names_ex_keys_datatypes.append(self.field_types[_curr_field_idx])
        
        
        _assignments = SqlList("ParameterAssignment")
        
        # Instantiate the assignments
                
        for _curr_field_idx in range(len(_field_names_ex_keys)):
            _left = ParameterIdentifier(_identifier = _field_names_ex_keys[_curr_field_idx])
            _right = ParameterParameter(_datatype = _field_names_ex_keys_datatypes[_curr_field_idx])
            _assignments.append(ParameterAssignment(_left, _right))


        # Create the WHERE conditions.

        _conditions = ParameterConditions()
        
        for _curr_field_idx in range(len(_key_field_names)):
            _left = ParameterIdentifier(_identifier = _key_field_names[_curr_field_idx])
            _right = ParameterParameter(_datatype = _key_field_datatypes[_curr_field_idx])
            _conditions.append(ParameterCondition( _left, _right, _operator = "="))
        
        # Specify target table
            
        _table_identifier =  ParameterIdentifier(_identifier = self.table_name)
        
        # Create VerbUpdate instance with all parameters
        
        _update = VerbUpdate(_table_identifier = _table_identifier, _conditions = _conditions, _assignments = _assignments)
        
        # Generate the SQL with all parameter place holders
        
        self._structure_update_sql = _update.as_sql(self.dal.db_type)        
       
   
             
    def _structure_insert_row(self, _row_idx, _row_data, _commit=True, _no_logging=None):
        if _commit is True:
            """Override parent to add SQL handling"""
            _execute_many_data = self._extract_data_columns_from_diff_row(range(len(self.field_names)), _row_data)

            # Apply and commit changes to the structure

            self.dal.executemany(self._structure_insert_sql, [_execute_many_data])
            if not _no_logging:
                self.log_insert("N/A in RDBMS", _row_data, "Destination table: " + self.table_name)
        # Call parent
        super(RDBMSDataset, self)._structure_insert_row(_row_idx,_row_data, _no_logging=_commit)
        
    def _structure_update_row(self, _row_idx, _row_data, _commit=True, _no_logging=None):
        if _commit is True:
            """Override parent to add SQL handling"""
            # To satisfy the VerbUpdate instance, create a two-dimensional array, leftmost columns are data, rightmost are keys.

            _field_idx_ex_keys = list(set(range(len(self.field_names))) - set(self._structure_key_fields))
            _execute_data = self._extract_data_columns_from_diff_row(_field_idx_ex_keys + self._structure_key_fields, _row_data)

            # Apply and commit changes to the database
            self.dal.executemany(self._structure_update_sql,  [_execute_data])
            if not _no_logging:
                self.log_update_row(_row_idx, self.data_table[_row_idx], _row_data, "Destination table: " + self.table_name)
        # Call parent
        super(RDBMSDataset, self)._structure_update_row(_row_idx,_row_data, _no_logging=_commit)

    def _structure_delete_row(self, _row_idx, _commit=True, _no_logging=None):
        if _commit is True:
            """Override parent to add SQL handling"""

            # Extract the key data
            _key_values = self._extract_data_columns_from_diff_row(self._structure_key_fields, self.data_table[_row_idx])
            # Make the deletes
            self.dal.executemany(self._structure_delete_sql, [_key_values])
            if not _no_logging:
                self.log_delete(_key_values, self.data_table[_row_idx], "Destination table: " + self.table_name)

        # Call parent
        super(RDBMSDataset, self)._structure_delete_row(_row_idx, _no_logging=_commit)
        #self.data_table.pop(_row_idx)
                
    def load(self):
        if self.table_name and self.table_name != "":
            """Query all values from a table from a RDBMS resource"""
            if not self.dal.connected:
                self.dal.connect_to_db()

            self.data_table = self.dal.query(select_all_skeleton(self.table_name, self.field_names).as_sql(self.dal.db_type))
            self.field_names = self.dal.field_names
            self.field_types = self.dal.field_types
        else:
            raise Exception("RDBMSDataset.load(): data_table must be set.")
        
        return self.data_table
    
    def save (self):
        """Save data. Commits the transaction"""
        self.dal.commit()
    


"""
Created on Nov 3, 2013

@author: Nicklas Boerjesson
"""

from lxml import etree
from os.path import exists

from qal.common.resources import Resources, Resource
from qal.common.strings import string_to_bool, make_path_absolute
from qal.common.mapping import Mapping

from qal.common.transform import perform_transformations, IfEmpty, Replace
from qal.dataset.factory import dataset_from_resource

from qal.common.xml_utils import xml_isnone
from qal.dataset.rdbms import RDBMSDataset
from qal.sql.macros import create_table_skeleton
from qal.tools.meta_queries import Meta_Queries


class Merge(object):
    """
    The merge class takes two datasets and merges them together.
    """
    mappings = None
    """A list of Mapping(qal.common.mapping) instances"""
    key_fields = None
    """A list of key field field indexes (in the destination dataset)"""
    source = None
    """The source dataset"""
    destination = None
    """The destination dataset"""
    resources = None
    """A instance of a Resources object(see qal.common.resources)"""
    destination_log_level = None
    """The log level to pass to the destination dataset"""
    post_execute_sql = None
    """An optional SQL that is exececute in the destination context, only applicable to RDBMS destinations"""
    insert = None
    """If data that exists in the source should be inserted into the destination."""
    delete = None
    """If data that do not exist in the source should be deleted the destination."""
    update = None
    """If rows that are different in the source should be updated in the destination"""

    def __init__(self, _xml_node=None):
        """
        Constructor
        """
        self.mappings = []
        self.key_fields = []
        if _xml_node != None:
            self.load_from_xml_node(_xml_node)

    def _field_mappings_as_xml_node(self):
        _xml_node = etree.Element("field_mappings")
        for _curr_mapping in self.mappings:
            _xml_node.append(_curr_mapping.as_xml_node())

        return _xml_node

    def _mappings_as_xml_node(self):
        _xml_node = etree.Element("mappings")
        _xml_node.append(self._field_mappings_as_xml_node())
        return _xml_node

    def _settings_as_xml_node(self):
        _xml_node = etree.Element("settings")
        etree.SubElement(_xml_node, "insert").text = str(self.insert)
        etree.SubElement(_xml_node, "update").text = str(self.update)
        etree.SubElement(_xml_node, "delete").text = str(self.delete)
        etree.SubElement(_xml_node, "post_execute_sql").text = self.post_execute_sql
        return _xml_node


    def as_xml_node(self):
        _xml_node = etree.Element('merge')
        _xml_node.append(self._mappings_as_xml_node())
        _xml_node.append(self._settings_as_xml_node())

        if self.resources is None:
            self.resources = Resources()

        if self.source is not None:
            try:
                _source_resource = self.resources.get_resource('source_uuid')
            except:
                _source_resource = None
            if _source_resource is None:
                _source_resource = Resource()
                _source_resource.uuid = 'source_uuid'
                _source_resource.caption = "source"
                self.resources.local_resources['source_uuid'] = _source_resource
            self.source.write_resource_settings(_source_resource)
        if self.destination is not None:
            try:
                _dest_resource = self.resources.get_resource('dest_uuid')
            except:
                _dest_resource = None
            if _dest_resource is None:
                _dest_resource = Resource()
                _dest_resource.uuid = 'dest_uuid'
                _dest_resource.caption = "destination"
                self.resources.local_resources['dest_uuid'] = _dest_resource

            self.destination.write_resource_settings(_dest_resource)

        """ TODO: Handle remotely defined resources properly. This way, they are included in the XML, which perhaps isn't
            right. Perhaps get_resource should return a tuple with a source parameter.
        """

        if self.source is not None or self.destination is not None:
            # If either aren't set, anything in resources are likely to be residuals from earlier.
            # However, there could be old resources left in one of them.
            _xml_node.append(self.resources.as_xml_node())
        return _xml_node

    def load_field_mappings_from_xml_node(self, _xml_node):
        if _xml_node != None:
            _mapping_idx = 0
            for _curr_mapping in _xml_node.findall("field_mapping"):
                _new_mapping = Mapping(_xml_node=_curr_mapping)
                self.mappings.append(_new_mapping)
                if _new_mapping.is_key == True:
                    self.key_fields.append(_mapping_idx)
                _mapping_idx += 1
        else:
            raise Exception("Merge.load_field_mappings_from_xml_node: Missing 'field_mappings'-node.")

    def load_mappings_from_xml_node(self, _xml_node):
        if _xml_node != None:
            self.load_field_mappings_from_xml_node(_xml_node.find("field_mappings"))
        else:
            raise Exception("Merge.load_field_mappings_from_xml_node: Missing 'mappings'-node.")


    def load_settings_from_xml_node(self, _xml_node):
        if _xml_node != None:
            self.insert = string_to_bool(xml_isnone(_xml_node.find("insert")))
            self.update = string_to_bool(xml_isnone(_xml_node.find("update")))
            self.delete = string_to_bool(xml_isnone(_xml_node.find("delete")))
            self.post_execute_sql = xml_isnone(_xml_node.find("post_execute_sql"))
        else:
            raise Exception("Merge.load_settings_from_xml_node: Missing 'settings'-node.")

    def load_from_xml_node(self, _xml_node):

        if _xml_node != None:
            self.load_mappings_from_xml_node(_xml_node.find("mappings"))
            self.load_settings_from_xml_node(_xml_node.find("settings"))
            self.resources = Resources(_resources_node=_xml_node.find("resources"))
            self.source = dataset_from_resource(self.resources.get_resource('source_uuid'))
            self.destination = dataset_from_resource(self.resources.get_resource('dest_uuid'))

        else:
            raise Exception("Merge.load_from_xml_node: \"None\" is not a valid Merge node.")


    def _mappings_to_fields(self, _dataset, _use_dest=True):

        _dataset.field_names = []
        _dataset.field_types = []
        self.key_fields = []
        if hasattr(_dataset, "field_xpaths"):
            _dataset.field_xpaths = []

        for _curr_mapping_idx in range(len(self.mappings)):
            _curr_mapping = self.mappings[_curr_mapping_idx]
            if _use_dest:
                _curr_source_ref = _curr_mapping.dest_reference
            else:
                _curr_source_ref = _curr_mapping.src_reference
            if _curr_source_ref is not None and _curr_source_ref != "":
                _dataset.field_names.append(_curr_source_ref)
                if hasattr(_dataset, "filename"):
                    _dataset.field_types.append("string")
                    if hasattr(_dataset, "field_xpaths"):
                        _dataset.field_xpaths.append(_curr_source_ref)


    def _load_datasets(self):

        # Load source_dataset
        try:
            self.source.load()
        except Exception as e:
            raise Exception("Merge._load_datasets: Failed loading data for source data set.\n" + \
                            "Check your mappings and other settings.\n" + \
                            "Dataset: " + str(self.source.__class__.__name__) + "\n" + \
                            "Error: " + str(e))
        if self.source.field_names is None or len(self.source.field_names) == 0:
            self._mappings_to_fields(self.source, False)




        # Load destination dataset
        try:
            # Handle non-existing destination files or tables
            if (hasattr(self.destination, "filename") and
                    not exists( make_path_absolute(self.destination.filename,self.destination._base_path))) or \
                    (isinstance(self.destination, RDBMSDataset) and
                        len(Meta_Queries.table_info(self.destination.dal, self.destination.table_name)) == 0):
                self.destination.field_names = [_curr_mapping.dest_reference for _curr_mapping in self.mappings]
                self.destination.data_table = []

            else:
                self.destination.load()
        except Exception as e:
            raise Exception("Merge._load_datasets: Failed loading data for destination data set.\n" + \
                            "Check your mappings and other settings.\n" + \
                            "Dataset: " + str(self.destination.__class__.__name__) + "\n" + \
                            "Error: " + str(e))

        if self.destination.field_names is None or len(self.destination.field_names) == 0:
            self._mappings_to_fields(self.destination, True)

        if self.destination_log_level:
            self.destination._log_level = self.destination_log_level


    def _make_shortcuts_readd_keys(self, _mappings, _source_dataset, _destination_dataset):
        """Make a list of which source column index maps to which destination column index and readd keys"""
        _shortcuts = []
        _key_fields = []

        # Make mapping
        for _curr_mapping in _mappings:
            if _curr_mapping.src_reference is not None and _curr_mapping.src_reference != "":
                _src_idx = _source_dataset.field_names.index(_curr_mapping.src_reference)
                if _curr_mapping.is_key:
                    _key_fields.append(_src_idx)
            else:
                _src_idx = None

            _dest_idx = _destination_dataset.field_names.index(_curr_mapping.dest_reference)
            _shortcuts.append([_src_idx, _dest_idx, _curr_mapping])

        return _shortcuts, _key_fields



    def _remap(self, _shortcuts, _key_fields, _source_dataset, _destination_dataset):
        """Create a remapped source data set that has the same data in the same columns as the destination data set.
        Also  and remaps keys."""

        _mapped_source = []
        # Loop all rows in the source data set
        for _curr_row in _source_dataset.data_table:
            # Create an empty row with None-values to fill later
            _curr_mapped = []
            _curr_mapped.extend(None for x in _destination_dataset.field_names)


            # Loop all the shortcuts to remap the data from the source structure into the destinations 
            # structure while applying transformations.
            for _curr_shortcut in _shortcuts:
                # Set the correct field in the destination data set

                if _curr_shortcut[0] is not None:
                    _curr_mapped[_curr_shortcut[1]]  = _curr_row[_curr_shortcut[0]]
                else:
                    # The destination column did not exist in the source? Fill with None for now.
                    _curr_mapped[_curr_shortcut[1]]  = None

            _mapped_source.append(_curr_mapped)

        # Remap keys to match the fields in _mapped_source
        _mapped_keys = []
        for _curr_key_field in _key_fields:
            for _curr_shortcut in _shortcuts:
                if _curr_key_field == _curr_shortcut[0]:
                    _mapped_keys.append(_curr_shortcut[1])

        return _mapped_source, _mapped_keys

    def clear_log(self):
        if self.destination is not None:
            self.destination.clear_log()


    def set_max_identities_for_mappings(self, _shortcuts, _source_dataset, _destination_dataset):
        """Loop through all mapping that uses identity """

        def if_empty(_value):
            if _value == "":
                return 0
            else:
                try:
                    return int(_value)
                except ValueError:
                    return 0

        def find_max(_curr_shortcut):

            if _curr_shortcut[1] is not None:
                if len(_destination_dataset.data_table) > 0:
                    _dest_max = max([if_empty(x[_curr_shortcut[1]]) for x in _destination_dataset.data_table])
                else:
                    _dest_max = 0
                if _curr_shortcut[0] is not None:
                    _src_max = max([if_empty(x[_curr_shortcut[0]]) for x in _source_dataset.data_table])
                    if _src_max > _dest_max:
                        return _src_max

                return _dest_max
            else:
                return 0

        _result = {}
        for _curr_shortcut in _shortcuts:
            _curr_mapping = _curr_shortcut[2]
            for _curr_transformation in _curr_mapping.transformations:
                if (isinstance(_curr_transformation, IfEmpty) and
                                _curr_transformation.value is not None and
                                _curr_transformation.value == "::identity::") \
                    or (isinstance(_curr_transformation, Replace) and
                                _curr_transformation.new is not None and
                                _curr_transformation.new.find("::identity::") > 1):

                    _curr_mapping.substitution.set_identity(find_max(_curr_shortcut) + 1)
                    break

    def apply_modifications(self, _shortcuts, _data_table):


        _transformation_shortcuts = []
        # Make a list of columns with transformations
        for _curr_shortcut in _shortcuts:
            _curr_mapping = _curr_shortcut[2]
            if len(_curr_mapping.transformations) > 0:
                _transformation_shortcuts.append(_curr_shortcut)

        for _curr_row in _data_table:
            for _curr_shortcut in _transformation_shortcuts:
                _curr_mapping = _curr_shortcut[2]
                try:
                    _curr_row[_curr_shortcut[1]] = perform_transformations(_curr_row[_curr_shortcut[1]], _curr_mapping.transformations)
                except Exception as e:
                    raise Exception("Transformation:\nError in applying transformations for row " +
                                    str(_data_table.index(_curr_row)) + ", column \"" + str(_curr_mapping.dest_reference) +
                                    "\":\n" + str(e))

        return _data_table

    def execute(self, _commit=True):
        """
        Execute the merge and return the results.
        :param _commit: Actually save the result
        :return: The merged dataset, the destination log, deletes, inserts, updates
        """


        # Load both source and destination data tables
        self._load_datasets()

        # Make shortcuts between source and destination and add key fields
        _shortcuts, self.key_fields = self._make_shortcuts_readd_keys(self.mappings, self.source, self.destination)

        # Create a remapped version of the source dataset data table to match the field order of the destination
        _mapped_source, _mapped_keys = self._remap(_shortcuts, self.key_fields, self.source, self.destination)

        # If there are any identity references, find and set identity seed to the highest value in
        # either source or destination. If there aren't any, it will start att zero.
        self.set_max_identities_for_mappings(_shortcuts, self.source, self.destination)

        # Apply substitutions, transformations
        _mapped_source = self.apply_modifications(_shortcuts, _mapped_source)


        #Merge the datasets
        _merged_dataset, _deletes, _inserts, _updates = self.destination.apply_new_data(_mapped_source, _mapped_keys,
                                                                                        _insert=self.insert,
                                                                                        _update=self.update,
                                                                                        _delete=self.delete,
                                                                                        _commit=_commit)

        if _commit:
            self.destination.save()

            if self.post_execute_sql is not None and self.post_execute_sql != "" and hasattr(self.destination, "dal"):
                self.destination.dal.execute(self.post_execute_sql)
                self.destination.dal.commit()

        return _merged_dataset, self.destination._log, _deletes, _inserts, _updates

